# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@users.sourceforge.net>
# Michael Hale Ligh <michael.hale@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""These are various utilities for rekall."""
import __builtin__
import bisect
import cPickle
import cStringIO
import importlib
import itertools
import json
import ntpath
import re
import shutil
import socket
import tempfile
import threading
import time


def SmartStr(string, encoding="utf8"):
    """Forces the string to be an encoded byte string."""
    if type(string) == unicode:
        return string.encode("utf8", "ignore")

    try:
        return string.__unicode__().encode(encoding)
    except AttributeError:
        return str(string)


def SmartUnicode(string, encoding="utf8"):
    """Forces the string into a unicode object."""
    try:
        # Allow the object to have a __unicode__ method.
        return unicode(string)
    except UnicodeError:
        return str(string).decode(encoding, "ignore")


def Hexdump(data, width=16):
    """Hexdump function shared by various plugins """
    for offset in xrange(0, len(data), width):
        row_data = data[offset:offset + width]
        translated_data = [
            x if ord(x) < 127 and ord(x) > 32 else "." for x in row_data]
        hexdata = " ".join(["{0:02x}".format(ord(x)) for x in row_data])

        yield offset, hexdata, translated_data


def WriteHexdump(renderer, data, base=0, width=16):
    """Write the hexdump to the fd."""
    renderer.table_header([dict(name='Offset', cname="offset", style="address"),
                           dict(name='Hex', cname="hex", width=width * 3),
                           dict(name='Data', cname='data', width=width)])

    for offset, hexdata, translated_data in Hexdump(data):
        renderer.table_row(base + offset, hexdata, "".join(translated_data))


# This is a synchronize decorator.
def Synchronized(f):
    """Synchronization decorator."""

    def NewFunction(self, *args, **kw):
        if self.lock:
            with self.lock:
                return f(self, *args, **kw)
        else:
            return f(self, *args, **kw)

    return NewFunction


class Node(object):
    """An entry to a linked list."""
    next = None
    prev = None
    data = None

    def __init__(self, data):
        self.data = data

    def __str__(self):
        return "Node:" + SmartStr(self.data)


class LinkedList(object):
    """A simple doubly linked list used for fast caches."""

    def __init__(self):
        # We are the head node.
        self.next = self.prev = self
        self.size = 0
        self.lock = threading.RLock()

    def Append(self, data):
        return self.AppendNode(Node(data))

    def AppendNode(self, node):
        self.size += 1
        last_node = self.prev

        last_node.next = node
        node.prev = last_node
        node.next = self
        self.prev = node

        return node

    def PopLeft(self):
        """Returns the head node and removes it from the list."""
        if self.next is self:
            raise IndexError("Pop from empty list.")

        first_node = self.next
        self.Unlink(first_node)

        return first_node.data

    def Pop(self):
        """Returns the tail node and removes it from the list."""
        if self.prev is self:
            raise IndexError("Pop from empty list.")

        last_node = self.tail
        self.Unlink(last_node)

        return last_node.data

    def Unlink(self, node):
        """Removes a given node from the list."""
        self.size -= 1

        node.prev.next = node.next
        node.next.prev = node.prev
        node.next = node.prev = None

    def __iter__(self):
        p = self.next
        while p is not self:
            yield p.data
            p = p.next

    def __len__(self):
        return self.size

    def __str__(self):
        p = self.next
        s = []
        while p is not self:
            s.append(str(p.data))
            p = p.next

        return "[" + ", ".join(s) + "]"


class FastStore(object):
    """This is a cache which expires objects in oldest first manner.

    This implementation first appeared in PyFlag and refined in GRR.

    This class implements an LRU cache which needs fast updates of the LRU order
    for random elements. This is implemented by using a dict for fast lookups
    and a linked list for quick deletions / insertions.
    """

    def __init__(self, max_size=10, kill_cb=None, lock=False):
        """Constructor.

        Args:
             max_size: The maximum number of objects held in cache.
             kill_cb: An optional function which will be called on each
                                object terminated from cache.
             lock: If True this cache will be thread safe.
        """
        self._age = LinkedList()
        self._hash = {}
        self._limit = max_size
        self._kill_cb = kill_cb
        self.lock = None
        if lock:
            self.lock = threading.RLock()
        self.hits = self.misses = 0

    @Synchronized
    def Expire(self):
        """Expires old cache entries."""
        while len(self._age) > self._limit:
            x = self._age.PopLeft()
            self.ExpireObject(x)

    @Synchronized
    def Put(self, key, item):
        """Add the object to the cache."""
        hit = self._hash.get(key, self)
        if hit is not self:
            self._age.Unlink(hit[0])

        node = self._age.Append(key)
        self._hash[key] = (node, item)

        self.Expire()

        return key

    @Synchronized
    def ExpireObject(self, key):
        """Expire a specific object from cache."""
        _, item = self._hash.pop(key, (None, None))

        if self._kill_cb and item is not None:
            self._kill_cb(item)

        return item

    @Synchronized
    def ExpireRegEx(self, regex):
        """Expire all the objects with the key matching the regex."""
        reg = re.compile(regex)
        for key in self._hash.keys():
            if reg.match(key):
                self.ExpireObject(key)

    @Synchronized
    def ExpirePrefix(self, prefix):
        """Expire all the objects with the key having a given prefix."""
        for key in self._hash.keys():
            if key.startswith(prefix):
                self.ExpireObject(key)

    @Synchronized
    def Get(self, key):
        """Fetch the object from cache.

        Objects may be flushed from cache at any time. Callers must always
        handle the possibility of KeyError raised here.

        Args:
            key: The key used to access the object.

        Returns:
            Cached object.

        Raises:
            KeyError: If the object is not present in the cache.
        """
        hit = self._hash.get(key, self)
        if hit is self:
            self.misses += 1
            raise KeyError(key)

        # Remove the item and put to the end of the age list
        node, item = hit
        self._age.Unlink(node)
        self._age.AppendNode(node)
        self.hits += 1

        return item

    def __iter__(self):
        return self._hash.items()

    @Synchronized
    def __contains__(self, key):
        result = key in self._hash
        if result:
            node, _ = self._hash[key]
            self._age.Unlink(node)
            self._age.AppendNode(node)

        return result

    @Synchronized
    def __getitem__(self, key):
        return self.Get(key)

    @Synchronized
    def Flush(self):
        """Flush all items from cache."""
        while self._age:
            x = self._age.PopLeft()
            self.ExpireObject(x)

        self._hash = {}

    @Synchronized
    def __getstate__(self):
        """When pickled the cache is fushed."""
        if self._kill_cb:
            raise RuntimeError("Unable to pickle a store with a kill callback.")

        self.Flush()
        return dict(max_size=self._limit)

    def __setstate__(self, state):
        self.__init__(max_size=state["max_size"])


class AgeBasedCache(FastStore):
    """A cache which removes objects after some time."""

    def __init__(self, max_age=20, **kwargs):
        super(AgeBasedCache, self).__init__(**kwargs)
        self.max_age = max_age

    def Put(self, key, item):
        super(AgeBasedCache, self).Put(key, (item, time.time()))

    def Get(self, key):
        item, timestamp = super(AgeBasedCache, self).Get(key)

        if timestamp + self.max_age > time.time():
            return item

        else:
            self.ExpireObject(key)
            raise KeyError("Item too old.")


# Compensate for Windows python not supporting socket.inet_ntop and some
# Linux systems (i.e. OpenSuSE 11.2 w/ Python 2.6) not supporting IPv6.

def inet_ntop(address_family, packed_ip):

    def inet_ntop4(packed_ip):
        if not isinstance(packed_ip, str):
            raise TypeError("must be string, not {0}".format(type(packed_ip)))

        if len(packed_ip) != 4:
            raise ValueError("invalid length of packed IP address string")

        return "{0}.{1}.{2}.{3}".format(*[ord(x) for x in packed_ip])

    def inet_ntop6(packed_ip):
        if not isinstance(packed_ip, str):
            raise TypeError("must be string, not {0}".format(type(packed_ip)))

        if len(packed_ip) != 16:
            raise ValueError("invalid length of packed IP address string")

        words = []
        for i in range(0, 16, 2):
            words.append((ord(packed_ip[i]) << 8) | ord(packed_ip[i + 1]))

        # Replace a run of 0x00s with None
        numlen = [(k, len(list(g))) for k, g in itertools.groupby(words)]
        max_zero_run = sorted(sorted(
            numlen, key=lambda x: x[1], reverse=True),
                              key=lambda x: x[0])[0]
        words = []
        for k, l in numlen:
            if (k == 0) and (l == max_zero_run[1]) and not None in words:
                words.append(None)
            else:
                for i in range(l):
                    words.append(k)

        # Handle encapsulated IPv4 addresses
        encapsulated = ""
        if (words[0] is None) and (len(words) == 3 or (
                len(words) == 4 and words[1] == 0xffff)):
            words = words[:-2]
            encapsulated = inet_ntop4(packed_ip[-4:])
        # If we start or end with None, then add an additional :
        if words[0] is None:
            words = [None] + words
        if words[-1] is None:
            words += [None]
        # Join up everything we've got using :s
        return (":".join(
            ["{0:x}".format(w) if w is not None else "" for w in words]) +
                encapsulated)

    if address_family == socket.AF_INET:
        return inet_ntop4(packed_ip)
    elif address_family == socket.AF_INET6:
        return inet_ntop6(packed_ip)
    raise socket.error("[Errno 97] Address family not supported by protocol")


def ConditionalImport(name):
    try:
        return importlib.import_module(name)
    except ImportError:
        pass


# This is only available on unix systems.
fcntl = ConditionalImport("fcntl")


class FileLock(object):
    """A self releasing file lock."""

    def __init__(self, fd):
        self.fd = fd

    def __enter__(self):
        if fcntl:
            fcntl.flock(self.fd.fileno(), fcntl.LOCK_EX)
        return self.fd

    def __exit__(self, exc_type, exc_value, traceback):
        if fcntl:
            fcntl.flock(self.fd.fileno(), fcntl.LOCK_UN)


class TempDirectory(object):
    """A self cleaning temporary directory."""

    def __enter__(self):
        self.name = tempfile.mkdtemp()

        return self.name

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.name, True)


class AttributedString(object):
    """This is just a container for a string and some metadata."""
    highlights = None

    def __init__(self, value, highlights=None):
        self.highlights = highlights
        self.value = value

    def __unicode__(self):
        return unicode(self.value)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return "AttributedString(%s, highlights=%s)" % (repr(self.value),
                                                        repr(self.highlights))


class HexDumpedString(AttributedString):
    """A string which should be hex dumped."""


class AttributeDict(dict):
    """A dict that can be accessed via attributes."""
    dirty = False

    def __setattr__(self, attr, value):
        try:
            # Check that the object itself has this attribute.
            object.__getattribute__(self, attr)

            return object.__setattr__(self, attr, value)
        except AttributeError:
            self.Set(attr, value)

    def Get(self, item, default=None):
        return self.get(item, default)

    def Set(self, attr, value):
        self.dirty = True

        # Setting a key to None means to remove it from the cache. NOTE! This
        # must be exactly None not a NoneObject() since it should be possible to
        # cache a NoneObject() as the value of some key.
        if value is None:
            self.pop(attr, None)
        else:
            self[attr] = value

    def __getattr__(self, attr):
        # Do not allow private attributes to be set.
        if attr.startswith("_"):
            raise AttributeError(attr)

        return self.get(attr)

    def __dir__(self):
        return sorted(self)


def FormatIPAddress(family, value):
    """Formats a value as an ascii IP address determined by family."""
    if value == None:
        return value

    return socket.inet_ntop(
        getattr(socket, str(family)),
        value.obj_vm.read(value.obj_offset, value.obj_size))


def ntoh(value):
    size = value.obj_size
    if size == 2:
        return socket.ntohs(value.v())
    elif size == 4:
        return socket.ntohl(value.v())

    from rekall import obj
    return obj.NoneObject("Not a valid integer")


def Invert(dictionary):
    """Inverts keys and values in dictionary.

    Assume the keys and values are unique.
    """
    return {v:k for k, v in dictionary.items()}


def PPrint(data, depth=0):
    """A pretty printer for a profile.

    This only supports dict, list and non-unicode strings.

    This produces both a valid json and a valid python file.
    """
    result = []
    if type(data) is bool:
        return str(data).lower()

    if isinstance(data, dict):
        # Empty dicts emitted on one line.
        if not data:
            return "{}"

        result.append("{")
        tmp = []
        for key, value in sorted(data.items()):
            # Only emit non-empty dicts.
            if value != {}:
                tmp.append(
                    " %s%s: %s" % (
                        " " * depth,
                        json.dumps(str(key)),
                        PPrint(data[key], depth + 1).strip()))

        result.append(", \n".join(tmp)[depth:])

        result.append(" }")
        return "\n".join([(" " * depth + x) for x in result])

    if isinstance(data, (list, tuple)):
        for item in data:
            result.append(PPrint(item, depth).strip())

        res = "[" + ", ".join(result) + "]"

        return res

    if isinstance(data, basestring):
        return json.dumps(data)

    # JSON encodes None as null.
    elif data is None:
        return "null"

    return SmartStr(data)


DEFINE_REGEX = re.compile(r"#define\s+([A-Z0-9_]+)\s+((0x)?[0-9A-Z]+)")

def MaskMapFromDefines(text):
    """Generates a maskmap dict from a list of #defines.

    This function allows us to copy the relevant #define sections from header
    files without needing to manually edit them. We get to keep the comments etc
    for readability.
    """
    result = {}
    for line in text.splitlines():
        m = DEFINE_REGEX.search(line)
        if m:
            name = m.group(1)
            value = m.group(2)
            if m.group(3):
                value = int(value, 16)
            elif value.startswith("0"):
                value = int(value, 8)
            else:
                value = int(value)

            result[name] = value

    return result


def EnumerationFromDefines(text):
    """Generate an Enumeration from a list of #defines."""
    result = {}
    for line in text.splitlines():
        m = DEFINE_REGEX.search(line)
        if m:
            name = m.group(1)
            value = m.group(2)
            if m.group(3):
                value = int(value, 16)
            else:
                value = int(value)

            result[value] = name

    return result


# The following is from
# http://code.activestate.com/recipes/577197-sortedcollection/
# Created by Raymond Hettinger on Fri, 16 Apr 2010 (MIT)

class SortedCollection(object):
    '''Sequence sorted by a key function.

    SortedCollection() is much easier to work with than using bisect() directly.
    It supports key functions like those use in sorted(), min(), and max().
    The result of the key function call is saved so that keys can be searched
    efficiently.

    Instead of returning an insertion-point which can be hard to interpret, the
    five find-methods return a specific item in the sequence. They can scan for
    exact matches, the last item less-than-or-equal to a key, or the first item
    greater-than-or-equal to a key.

    Once found, an item's ordinal position can be located with the index()
    method.

    New items can be added with the insert() and insert_right() methods.
    Old items can be deleted with the remove() method.

    The usual sequence methods are provided to support indexing, slicing,
    length lookup, clearing, copying, forward and reverse iteration, contains
    checking, item counts, item removal, and a nice looking repr.

    Finding and indexing are O(log n) operations while iteration and insertion
    are O(n).  The initial sort is O(n log n).

    The key function is stored in the 'key' attibute for easy introspection or
    so that you can assign a new key function (triggering an automatic re-sort).

    In short, the class was designed to handle all of the common use cases for
    bisect but with a simpler API and support for key functions.

    >>> from pprint import pprint
    >>> from operator import itemgetter

    >>> s = SortedCollection(key=itemgetter(2))
    >>> for record in [
    ...         ('roger', 'young', 30),
    ...         ('angela', 'jones', 28),
    ...         ('bill', 'smith', 22),
    ...         ('david', 'thomas', 32)]:
    ...     s.insert(record)

    >>> pprint(list(s))         # show records sorted by age
    [('bill', 'smith', 22),
     ('angela', 'jones', 28),
     ('roger', 'young', 30),
     ('david', 'thomas', 32)]

    >>> s.find_le(29)           # find oldest person aged 29 or younger
    ('angela', 'jones', 28)
    >>> s.find_lt(28)           # find oldest person under 28
    ('bill', 'smith', 22)
    >>> s.find_gt(28)           # find youngest person over 28
    ('roger', 'young', 30)

    >>> r = s.find_ge(32)       # find youngest person aged 32 or older
    >>> s.index(r)              # get the index of their record
    3
    >>> s[3]                    # fetch the record at that index
    ('david', 'thomas', 32)

    >>> s.key = itemgetter(0)   # now sort by first name
    >>> pprint(list(s))
    [('angela', 'jones', 28),
     ('bill', 'smith', 22),
     ('david', 'thomas', 32),
     ('roger', 'young', 30)]

    '''

    def __init__(self, iterable=(), key=None):
        self._given_key = key
        key = (lambda x: x) if key is None else key
        decorated = sorted((key(item), item) for item in iterable)
        self._keys = [k for k, item in decorated]
        self._items = [item for k, item in decorated]
        self._key = key

    def _getkey(self):
        return self._key

    def _setkey(self, key):
        if key is not self._key:
            self.__init__(self._items, key=key)

    def _delkey(self):
        self._setkey(None)

    key = property(_getkey, _setkey, _delkey, 'key function')

    def clear(self):
        self.__init__([], self._key)

    def copy(self):
        return self.__class__(self, self._key)

    def __len__(self):
        return len(self._items)

    def __getitem__(self, i):
        return self._items[i]

    def __iter__(self):
        return iter(self._items)

    def __reversed__(self):
        return reversed(self._items)

    def __repr__(self):
        return '%s(%r, key=%s)' % (
            self.__class__.__name__,
            self._items,
            getattr(self._given_key, '__name__', repr(self._given_key))
        )

    def __reduce__(self):
        return self.__class__, (self._items, self._given_key)

    def __contains__(self, item):
        k = self._key(item)
        i = bisect.bisect_left(self._keys, k)
        j = bisect.bisect_right(self._keys, k)
        return item in self._items[i:j]

    def index(self, item):
        'Find the position of an item.  Raise ValueError if not found.'
        k = self._key(item)
        i = bisect.bisect_left(self._keys, k)
        j = bisect.bisect_right(self._keys, k)
        return self._items[i:j].index(item) + i

    def count(self, item):
        'Return number of occurrences of item'
        k = self._key(item)
        i = bisect.bisect_left(self._keys, k)
        j = bisect.bisect_right(self._keys, k)
        return self._items[i:j].count(item)

    def insert(self, item):
        'Insert a new item.  If equal keys are found, add to the left'
        k = self._key(item)
        i = bisect.bisect_left(self._keys, k)
        self._keys.insert(i, k)
        self._items.insert(i, item)

    def insert_right(self, item):
        'Insert a new item.  If equal keys are found, add to the right'
        k = self._key(item)
        i = bisect.bisect_right(self._keys, k)
        self._keys.insert(i, k)
        self._items.insert(i, item)

    def remove(self, item):
        'Remove first occurence of item.  Raise ValueError if not found'
        i = self.index(item)
        del self._keys[i]
        del self._items[i]

    def find(self, k):
        'Return first item with a key == k.  Raise ValueError if not found.'
        i = bisect.bisect_left(self._keys, k)
        if i != len(self) and self._keys[i] == k:
            return self._items[i]
        raise ValueError('No item found with key equal to: %r' % (k,))

    def find_le(self, k):
        'Return last item with a key <= k.  Raise ValueError if not found.'
        i = bisect.bisect_right(self._keys, k)
        if i:
            return self._items[i-1]
        raise ValueError('No item found with key at or below: %r' % (k,))

    def find_lt(self, k):
        'Return last item with a key < k.  Raise ValueError if not found.'
        i = bisect.bisect_left(self._keys, k)
        if i:
            return self._items[i-1]
        raise ValueError('No item found with key below: %r' % (k,))

    def find_ge(self, k):
        """Return first item with a key >= equal to k.

        Raise ValueError if not found
        """
        i = bisect.bisect_left(self._keys, k)
        if i != len(self):
            return self._items[i]
        raise ValueError('No item found with key at or above: %r' % (k,))

    def find_gt(self, k):
        'Return first item with a key > k.  Raise ValueError if not found'
        i = bisect.bisect_right(self._keys, k)
        if i != len(self):
            return self._items[i]
        raise ValueError('No item found with key above: %r' % (k,))

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self._keys == other._keys and self._items == other._items)


class RangedCollection(object):
    """A convenience wrapper around SortedCollection for ranges."""

    def __init__(self):
        self.collection = SortedCollection()

    def insert(self, start, end, data):
        self.collection.insert((int(start), int(end), data))

    def get_range(self, value):
        """Retrieve the data associated with the range that contains value."""
        try:
            start, end, data = self.collection.find_le((value+1, None, None))
            if start <= value < end:
                return data
        except ValueError:
            return None


class JITIterator(object):
    def __init__(self, baseclass):
        self.baseclass = baseclass

    def __contains__(self, item):
        return item in list(self)

    def __str__(self):
        return str(list(self))

    def __iter__(self):
        return (
            x.name for x in self.baseclass.classes.values() if x.name)


def CopyFDs(in_fd, out_fd, length=2**64):
    """Copy from one fd to another.

    If length is specified, we stop when we copied this many bytes. We always
    stop when in_fd reaches EOF.
    """
    while length > 0:
        data = in_fd.read(min(10000000, length))
        if not data:
            return

        out_fd.write(data)
        length -= len(data)


def CopyAStoFD(in_as, out_fd, start=0, length=2**64, cb=lambda x: None):
    """Copy an address space into a file-like object."""
    blocksize = 1024 * 1024

    for range_offset, _, range_length in in_as.get_available_addresses(
            start=start):
        range_end = range_offset + range_length

        for offset in xrange(range_offset, range_end, blocksize):
            to_read = min(blocksize, range_end - offset, length)
            data = in_as.read(offset, to_read)

            out_fd.seek(offset)
            out_fd.write(data)
            length -= len(data)

            cb(offset)


def issubclass(obj, cls):    # pylint: disable=redefined-builtin
    """A sane implementation of issubclass.

    See http://bugs.python.org/issue10569

    Python bare issubclass must be protected by an isinstance test first since
    it can only work on types and raises when provided something which is not a
    type.

    Args:
      obj: Any object or class.
      cls: The class to check against.

    Returns:
      True if obj is a subclass of cls and False otherwise.
    """
    return isinstance(obj, type) and __builtin__.issubclass(obj, cls)


def XOR(string1, string2):
    """Returns string1 xor string2."""
    return "".join(chr(ord(x) ^ ord(y)) for x, y in zip(string1, string2))


def xrange(start, end, step=1):
    """In Python2 the xrange builtin is broken.

    It raises when start or end do not fit in an int. Since python does not
    generally care about this we need to implement a clean version of this
    builtin.
    """
    x = start
    while x < end:
        yield x
        x += step


def SafePickle(data):
    """An unpickler for serialized tuple/lists/strings etc.

    Does not support recovering instances.
    """
    return cPickle.dumps(data, -1)

def SafeUnpickler(data):
    """An unpickler for serialized tuple/lists/strings etc.

    Does not support recovering instances.
    """
    unpickler = cPickle.Unpickler(cStringIO.StringIO(data))
    unpickler.find_global = None

    return unpickler.load()


def SplitPath(path):
    """Splits the path into a list of components."""
    result = []
    while 1:
        path, filename = ntpath.split(path)

        if filename:
            result.append(filename)
        else:
            if path and path not in ("\\", "/"):
                result.append(path)
            break

    result.reverse()
    return result


def CaseInsensitiveDictLookup(key, dictionary):
    """Lookup the dictionary with a case insensitive key."""
    # First try as is.
    result = dictionary.get(key)
    if result is None:
        for k, v in dictionary.iteritems():
            if k.lower() == key.lower():
                return v

    return result



# Rekall Memory Forensics
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Authors:
# Copyright (C) 2015 Michael Cohen <scudette@google.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""This Address Space allows us to open aff4 images.

AFF4 images are produced by the Rekall memory acquisition tools (Pmem and
friends).

For this address space to work:

pip install pyaff4

"""
import logging
import re
import os

from rekall import addrspace
from rekall import yaml_utils
from rekall import utils
from rekall.plugins.addrspaces import standard

from pyaff4 import data_store
from pyaff4 import zip
from pyaff4 import lexicon
from pyaff4 import rdfvalue

from pyaff4 import plugins  # pylint: disable=unused-import


# Control the logging level for the pyaff4 library logger.
LOGGER = logging.getLogger("pyaff4")
LOGGER.setLevel(logging.ERROR)


class AFF4StreamWrapper(object):
    def __init__(self, stream):
        self.stream = stream

    def read(self, offset, length):
        self.stream.seek(offset)
        return self.stream.read(length)

    def end(self):
        return self.stream.Size()

    def __unicode__(self):
        return utils.SmartUnicode(self.stream.urn)


class AFF4AddressSpace(addrspace.CachingAddressSpaceMixIn,
                       addrspace.MultiRunBasedAddressSpace):
    """Handle AFF4Map or AFF4Image type streams.

    Since AFF4 volumes may contain multiple streams, we allow the stream to be
    specified inside the volume path. For example suppose the volume located at:

    /home/mic/images/myimage.aff4

    Contains a stream called PhysicalMemory, then we can specify the filename
    as:

    /home/mic/images/myimage.aff4/PhysicalMemory

    If we just specified the path to the volume, then this address space will
    pick the first AFF4 stream which has an aff4:category of
    lexicon.AFF4_MEMORY_PHYSICAL.

    So if you have more than one physical memory stream in the same volume, you
    will need to specify the full path to the stream within the volume.
    """
    __name = "aff4"
    __image = True

    # This AS can map files into itself.
    __can_map_files = True

    order = standard.FileAddressSpace.order - 10

    def __init__(self, filename=None, **kwargs):
        super(AFF4AddressSpace, self).__init__(**kwargs)

        self.as_assert(self.base == None,
                       "Must stack on another address space")

        path = filename or self.session.GetParameter("filename")
        self.as_assert(path != None, "Filename must be specified")

        self.image = None
        self.phys_base = self
        self.resolver = data_store.MemoryDataStore()

        # A map between the filename and the offset it is mapped into the
        # address space.
        self.mapped_files = {}

        try:
            volume_path, stream_path = self._LocateAFF4Volume(path)
        except IOError:
            raise addrspace.ASAssertionError("Unable to open AFF4 volume")

        # filename is a volume, and there is no stream specified, just autoload
        # the stream if possible.
        if not stream_path:
            try:
                self._AutoLoadAFF4Volume(volume_path)
                return
            except IOError:
                raise addrspace.ASAssertionError("Unable to open AFF4 volume")

        # If the user asked for a specific stream just load that one. Note that
        # you can still load the pagefile manually using the --pagefile
        # parameter.
        with zip.ZipFile.NewZipFile(self.resolver, volume_path) as volume:
            self.volumes.add(volume.urn)
            self._LoadMemoryImage(volume.urn.Append(stream_path))

    def _LocateAFF4Volume(self, filename):
        stream_name = []
        volume_path = filename
        while volume_path:
            try:
                volume_urn = rdfvalue.URN.FromFileName(volume_path)
                with self.resolver.AFF4FactoryOpen(volume_urn) as aff4_stream:
                    if stream_name:
                        return aff4_stream.urn, os.path.join(*stream_name)

                    return aff4_stream.urn, None

            except IOError:
                volume_path, stream_component = os.path.split(volume_path)
                if not stream_component:
                    break

                stream_name.insert(0, stream_component)

        raise IOError("Not found")

    def _AutoLoadAFF4Volume(self, path):
        with zip.ZipFile.NewZipFile(self.resolver, path) as volume:
            self.volume_urn = volume.urn

            # We are searching for images with the physical memory category.
            for (subject, _, value) in self.resolver.QueryPredicate(
                    lexicon.AFF4_CATEGORY):
                if value == lexicon.AFF4_MEMORY_PHYSICAL:
                    self._LoadMemoryImage(subject)
                    break

        self.as_assert(self.image is not None,
                       "No physical memory categories found.")

        self.filenames = {}
        # Newer AFF4 images should have the AFF4_STREAM_ORIGINAL_FILENAME
        # attribute set.
        for (subject, _, value) in self.resolver.QueryPredicate(
                lexicon.AFF4_STREAM_ORIGINAL_FILENAME):
            # Normalize the filename for case insensitive filesysyems.
            self.filenames[unicode(value).lower()] = subject

        # TODO: Deprecate this guessing once all images have the
        # AFF4_STREAM_ORIGINAL_FILENAME attribute.
        for subject in self.resolver.QuerySubject(re.compile(".")):
            relative_name = self.volume_urn.RelativePath(subject)
            if relative_name:
                filename = self._normalize_filename(relative_name)
                self.filenames[filename] = subject

    def _normalize_filename(self, filename):
        """Normalize the filename based on the source OS."""
        m = re.match(r"/?([a-zA-Z]:[/\\].+)", filename)
        if m:
            # This is a windows filename.
            filename = m.group(1).replace("/", "\\")

            # The 32 bit WinPmem imager access native files via SysNative but
            # they are really located in System32.
            filename = filename.replace("SysNative", "System32")

            return filename.lower()

        return filename


    def _LoadMemoryImage(self, image_urn):
        aff4_stream = self.resolver.AFF4FactoryOpen(image_urn)
        self.image = AFF4StreamWrapper(aff4_stream)

        # Add the ranges if this is a map.
        try:
            for map_range in aff4_stream.GetRanges():
                self.runs.insert((map_range.map_offset,
                                  map_range.map_offset,
                                  map_range.length,
                                  self.image))
        except AttributeError:
            self.runs.insert((0, 0, aff4_stream.Size(), self.image))

        self._parse_physical_memory_metadata(aff4_stream.urn)
        self.session.logging.info("Added %s as physical memory", image_urn)

    def file_mapping_offset(self, filename):
        """Returns the offset where the filename should be mapped.

        This function manages the session cache. By storing the file mappings in
        the session cache we can guarantee repeatable mappings.
        """
        mapped_files = self.session.GetParameter("file_mappings", {})
        if filename in mapped_files:
            return utils.CaseInsensitiveDictLookup(
                filename, mapped_files)

        # Give a bit of space for the mapping and page align it.
        mapped_offset = (self.end() + 0x10000) & 0xFFFFFFFFFFFFF000
        mapped_files[filename] = mapped_offset

        self.session.SetCache("file_mappings", mapped_files)

        return mapped_offset

    def get_mapped_offset(self, filename, file_offset):
        """Map the filename into the address space.

        If the filename is found in the AFF4 image, we return the offset at
        which it is mapped. Otherwise return None.
        """
        mapped_offset = None
        filename = self._normalize_filename(filename)
        mapped_offset = utils.CaseInsensitiveDictLookup(
            filename, self.mapped_files)
        if mapped_offset is None:
            # Try to map the file.
            subject = utils.CaseInsensitiveDictLookup(
                filename, self.filenames)

            if subject:
                stream = self.resolver.AFF4FactoryOpen(subject)
                mapped_offset = self.file_mapping_offset(filename)
                self.add_run(mapped_offset, 0, stream.Size(),
                             AFF4StreamWrapper(stream))

                self.session.logging.info(
                    "Mapped %s into address %#x", stream.urn, mapped_offset)

            else:
                # Cache failures too.
                mapped_offset = -1

        # Cache for next time.
        self.mapped_files[filename] = mapped_offset
        if mapped_offset > 0:
            return mapped_offset + file_offset

    def _parse_physical_memory_metadata(self, image_urn):
        try:
            with self.resolver.AFF4FactoryOpen(
                    image_urn.Append("information.yaml")) as fd:
                metadata = yaml_utils.decode(fd.read(10e6))
                # Allow the user to override the AFF4 file.
                if not self.session.HasParameter("dtb"):
                    self.session.SetCache(
                        "dtb", metadata.get("Registers", {}).get("CR3"),
                        volatile=False)

                if not self.session.HasParameter("kernel_base"):
                    self.session.SetCache(
                        "kernel_base", metadata.get("KernBase"),
                        volatile=False)
        except IOError:
            self.session.logging.warn(
                "AFF4 volume does not contain PhysicalMemory metadata.")

    def describe(self, address):
        try:
            virt_addr, file_address, _, stream_as = self.runs.find_le(
                address)

            # For normal physical memory addresses just be concise.
            if stream_as == self.image:
                return u"%#x" % address

            # For other mapped streams, just say which ones they are.
            return u"%#x @ %s (Mapped %#x)" % (
                address - virt_addr + file_address,
                stream_as, address)

        except ValueError:
            pass

        # For unmapped streams just say we have no idea.
        return u"%#x (Unmapped)" % address



# pylint: disable=unused-import
# Add these so that pyinstaller builds these dependencies in.
import rdflib.plugins.memory
import rdflib.plugins.parsers.hturtle
import rdflib.plugins.parsers.notation3
import rdflib.plugins.parsers.nquads
import rdflib.plugins.parsers.nt
import rdflib.plugins.parsers.rdfxml
import rdflib.plugins.parsers.structureddata
import rdflib.plugins.parsers.trig
import rdflib.plugins.parsers.trix
import rdflib.plugins.serializers.n3
import rdflib.plugins.serializers.nquads
import rdflib.plugins.serializers.nt
import rdflib.plugins.serializers.rdfxml
import rdflib.plugins.serializers.trig
import rdflib.plugins.serializers.trix
import rdflib.plugins.serializers.turtle
import rdflib.plugins.sleepycat
import rdflib.plugins.sparql.processor
import rdflib.plugins.sparql.results.csvresults
import rdflib.plugins.sparql.results.jsonresults
import rdflib.plugins.sparql.results.tsvresults
import rdflib.plugins.sparql.results.txtresults
import rdflib.plugins.sparql.results.xmlresults
import rdflib.plugins.stores.auditable
import rdflib.plugins.stores.concurrent
import rdflib.plugins.stores.sparqlstore



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# The code in this directory is based on original code and algorithms by Andrew
# Case (atcuno@gmail.com).
__author__ = "Michael Cohen <scudette@google.com>"

import re

from rekall import kb
from rekall import obj
from rekall import plugin
from rekall import scan
from rekall import utils

from rekall.plugins import core

# A few notes on XNU's (64bit) memory layout:
#
# Because of the way the Darwin kernel (XNU) is bootstrapped, a section of its
# virtual address space maps linearly to the base of the physical address space.
# This relationship is basically:
# KERNEL_MIN_ADDRESS + the_physical_address = the_virtual_address
#
# The kernel ensures this when allocating certain data structures, most notably
# the page tables [1]. However, the kernel doesn't actually "know" the value of
# KERNEL_MIN_ADDRESS, which is defined the Makefile [2]. Instead, allocations
# are done by keeping a cursor at the lowest available physical address [3].
#
# Because of this, when the kernel needs to convert an address from the virtual
# address space to the physical address space without relying on the page
# tables, it uses a "safer" variation on the above rule and masks out the first
# 32 bits of the address using a macro called ID_MAP_VTOP [4,5], which is a
# simple bitmask (LOW_4GB_MASK [6]).
#
# We copy/adapt all three #defines below. When we need to bootstrap the virtual
# address space, relying on ID_MAP_VTOP is preferrable, because it's less
# fragile. However, KERNEL_MIN_ADDRESS can be a good heuristic for deciding
# whether a particular value is a valid pointer in the kernel virtual address
# space, so I decided to keep it around.
#
# [1]
# github.com/opensource-apple/xnu/blob/10.9/osfmk/i386/i386_init.c#L134
#
# [2]
# github.com/opensource-apple/xnu/blob/10.9/makedefs/MakeInc.def#L258
#
# [3] This is where physfree is defined as the next free page, after a blank
# page, after the last page of the kernel image as determined by the bootloader.
# github.com/opensource-apple/xnu/blob/10.9/osfmk/i386/i386_init.c#L330
#
# [4]
# github.com/opensource-apple/xnu/blob/10.9/osfmk/i386/pmap.h#L353
#
# [5] Example use, to set the physical address of the DTB when switching address
# spaces, knowing the virtual address of the first page table:
# github.com/opensource-apple/xnu/blob/10.9/osfmk/i386/pal_routines.c#L254
#
# [6]
# github.com/opensource-apple/xnu/blob/10.9/osfmk/i386/pmap.h#L119
LOW_4GB_MASK = 0x00000000ffffffff
KERNEL_MIN_ADDRESS = 0xffffff8000000000


def ID_MAP_VTOP(x):
    return x & LOW_4GB_MASK

# On x64, only 48 bits of the pointer are addressable.
X64_POINTER_MASK = 0x0000ffffffffffff


def MOUNTAIN_LION_OR_LATER(profile):
    return bool(profile.get_constant("_BootPML4", False))


class KernelSlideHook(kb.ParameterHook):
    """Find the kernel slide if needed."""

    name = "vm_kernel_slide"

    def calculate(self):
        if MOUNTAIN_LION_OR_LATER(self.session.profile):
            return DarwinFindKASLR(session=self.session).vm_kernel_slide()

        # Kernel slide should be treated as 0 if not relevant.
        return 0


class CatfishScanner(scan.BaseScanner):
    checks = [
        ("StringCheck", dict(needle="Catfish \x00\x00"))
    ]


class CatfishOffsetHook(kb.ParameterHook):
    """Find the actual offset of the _lowGlo struct."""

    name = "catfish_offset"

    def calculate(self):
        for hit in CatfishScanner(
                address_space=self.session.physical_address_space,
                session=self.session).scan():
            return hit


class DarwinKASLRMixin(object):
    """Ensures that KASLR slide is computed and stored in the session."""

    @classmethod
    def args(cls, parser):
        super(DarwinKASLRMixin, cls).args(parser)

        parser.add_argument("--vm_kernel_slide", type="IntParser",
                            help="OS X 10.8 and later: kernel ASLR slide.")

    def __init__(self, vm_kernel_slide=None, **kwargs):
        """A mixin for Darwin plugins that require a valid KASLR slide.

        Args:
          vm_kernel_slide: The integer KASLR slide used in this image. If not
          given it will be computed.
        """
        super(DarwinKASLRMixin, self).__init__(**kwargs)

        if not MOUNTAIN_LION_OR_LATER(self.profile):
            return

        if vm_kernel_slide is not None:
            self.session.SetCache("vm_kernel_slide", vm_kernel_slide)


class AbstractDarwinCommandPlugin(plugin.PhysicalASMixin,
                                  plugin.ProfileCommand):
    """A base class for all darwin based plugins."""
    __abstract = True

    @classmethod
    def is_active(cls, session):
        """We are only active if the profile is darwin."""
        return (session.profile.metadata("os") == "darwin" and
                plugin.Command.is_active(session))


class AbstractDarwinParameterHook(kb.ParameterHook):

    @classmethod
    def is_active(cls, session):
        """We are only active if the profile is Darwin."""
        return (super(AbstractDarwinParameterHook, cls).is_active(session) and
                session.profile.metadata("os") == 'darwin')


class DarwinFindKASLR(AbstractDarwinCommandPlugin):
    """A scanner for KASLR slide values in the Darwin kernel.

    The scanner works by looking up a known data structure and comparing
    its actual location to its expected location. Verification is a similar
    process, using a second constant. This takes advantage of the fact that both
    data structures are in a region of kernel memory that maps to the physical
    memory in a predictable way (see ID_MAP_VTOP).

    Human-readable output includes values of the kernel version string (which is
    used for validation) for manual review, in case there are false positives.
    """

    __name = "find_kaslr"

    @classmethod
    def is_active(cls, session):
        return (super(DarwinFindKASLR, cls).is_active(session) and
                MOUNTAIN_LION_OR_LATER(session.profile))

    def all_catfish_hits(self):
        """Yields possible lowGlo offsets, starting with session-cached one.

        Because the first hit on the catfish string isn't necessarily the right
        one, this function will yield subsequent ones by scanning the physical
        address space, starting with the offset of the cached first hit.

        The caller is responsible for updating the session cache with the
        correct offset.
        """
        first_hit = self.session.GetParameter("catfish_offset")
        yield first_hit

        for hit in CatfishScanner(
                address_space=self.session.physical_address_space,
                session=self.session).scan(offset=first_hit + 1):
            yield hit

    def vm_kernel_slide_hits(self):
        """Tries to compute the KASLR slide.

        In an ideal scenario, this should return exactly one valid result.

        Yields:
          (int) semi-validated KASLR value
        """

        expected_offset = self.profile.get_constant(
            "_lowGlo", is_address=False)
        expected_offset = ID_MAP_VTOP(expected_offset)

        for hit in self.all_catfish_hits():
            vm_kernel_slide = int(hit - expected_offset)

            if self._validate_vm_kernel_slide(vm_kernel_slide):
                self.session.SetCache("catfish_offset", hit)
                yield vm_kernel_slide

    def vm_kernel_slide(self):
        """Returns the first result of vm_kernel_slide hits and stops the scan.

        This is the idiomatic way of using this plugin if all you need is the
        likely KASLR slide value.

        Returns:
          A value for the KASLR slide that appears sane.
        """
        self.session.logging.debug("Searching for KASLR hits.")
        for vm_kernel_slide in self.vm_kernel_slide_hits():
            return vm_kernel_slide

    def _lookup_version_string(self, vm_kernel_slide):
        """Uses vm_kernel_slide to look up kernel version string.

        This is used for validation only. Physical address space is
        asumed to map to kernel virtual address space as expressed by
        ID_MAP_VTOP.

        Args:
          vm_kernel_slide: KASLR slide to be used for lookup. Overrides whatever
          may already be set in session.

        Returns:
          Kernel version string (should start with "Darwin Kernel"
        """
        version_offset = self.profile.get_constant(
            "_version", is_address=False)
        version_offset += vm_kernel_slide
        version_offset = ID_MAP_VTOP(version_offset)

        return self.profile.String(vm=self.physical_address_space,
                                   offset=version_offset)

    def _validate_vm_kernel_slide(self, vm_kernel_slide):
        """Checks sanity of vm_kernel_slide by looking up kernel version.
        If the result a string that looks like the kernel version string the
        slide value is assumed to be valid. Note that this can theoretically
        give false positives.

        Args:
          vm_kernel_slide: KASLR slide to be used for validation. Overrides
          whatever may already be set in session.

        Returns:
          True if vm_kernel_slide value appears sane. False otherwise.
        """
        version_string = self._lookup_version_string(vm_kernel_slide)
        return version_string[0:13] == "Darwin Kernel"

    def render(self, renderer):
        renderer.table_header([
            ("KASLR Slide", "vm_kernel_slide", "[addrpad]"),
            ("Kernel Version", "_version", "30"),
        ])

        for vm_kernel_slide in self.vm_kernel_slide_hits():
            renderer.table_row(vm_kernel_slide,
                               self._lookup_version_string(vm_kernel_slide))


class DarwinFindDTB(DarwinKASLRMixin, AbstractDarwinCommandPlugin,
                    core.FindDTB):
    """Tries to find the DTB address for the Darwin/XNU kernel.

    As the XNU kernel developed over the years, the best way of deriving this
    information changed. This class now offers multiple methods of finding the
    DTB. Calling find_dtb should automatically select the best method for the
    job, based on the profile. It will also attempt to fall back on less ideal
    ways of getting the DTB if the best way fails.
    """

    __name = "find_dtb"

    def _dtb_hits_idlepml4(self):
        """On 10.8 and later, x64, tries to determine the DTB using IdlePML4.

        IdlePML4 is the address (in Kernel AS) of the kernel DTB [1]. The DTB
        itself happens to be located in a section of kernel memory that sits at
        the base of the physical address space [2], and its virtual address can
        be converted to its physical address using the ID_MAP_VTOP macro
        which kernel defines for this express purpose [3].

        Should work on: 10.8 and later.
        Best for: 10.9 and later.

        Yields:
          The physical address of the DTB, not verified.

        1:
        github.com/opensource-apple/xnu/blob/10.9/osfmk/i386/i386_init.c#L281

        Here the kernel initializes the page register at the address IdlePML4
        points to (masked using the bitmask macro). The same function switches
        to the newly initialized address space right before returning.

        // IdlePML4 single entry for kernel space.
        fillkpt(IdlePML4 + KERNEL_PML4_INDEX,
                INTEL_PTE_WRITE, (uintptr_t)ID_MAP_VTOP(IdlePDPT), 0, 1);

        2:
        The first page of IdlePML4 is allocated by the ALLOCPAGES function
        located here:
        github.com/opensource-apple/xnu/blob/10.9/osfmk/i386/i386_init.c#L134

        3:
        ID_MAP_VTOP is defined here, as simple bitmask:
        github.com/opensource-apple/xnu/blob/10.9/osfmk/i386/pmap.h#L353
        """
        idlepml4 = ID_MAP_VTOP(self.profile.get_constant(
            "_IdlePML4", True))
        dtb = self.profile.Object("unsigned int", offset=idlepml4,
                                  vm=self.physical_address_space)
        yield int(dtb)

    def _dtb_hits_legacy(self):
        """The original way of getting the DTB, adapted from Volatility.

        I have no idea how or why this is intended to work, but it seems to for
        old images.

        Should work on: 10.7 and earlier.

        Yields:
          The physical address of the DTB, not verified.
        """
        if self.profile.metadata("arch") == "I386":
            result = self.profile.get_constant("_IdlePDPT")

            # Since the DTB must be page aligned, if this is not, it is probably
            # a pointer to the real DTB.
            if result % 0x1000:
                result = self.profile.get_constant_object(
                    "_IdlePDPT", "unsigned int")

            yield result
        else:
            result = self.profile.get_constant("_IdlePML4", is_address=True)
            if result > 0xffffff8000000000:
                result -= 0xffffff8000000000

            yield result

    def _dtb_hits_kernel_pmap(self):
        """On 64-bit systems, finds the DTB from the kernel pmap struct.

        This is a very easy way of getting the DTB on systems where the kernel
        pmap is a static symbol (which seems to be most of them.)

        Yields:
          The physical address of the DTB, not verified.
        """
        kernel_pmap_addr = self.profile.get_constant(
            "_kernel_pmap_store", is_address=True)
        kernel_pmap = self.profile.pmap(offset=ID_MAP_VTOP(kernel_pmap_addr),
                                        vm=self.physical_address_space)
        yield int(kernel_pmap.pm_cr3)

    def _dtb_methods(self):
        """Determines viable methods of getting the DTB based on profile.

        Yields:
          Callable object that will yield DTB values.
        """
        if MOUNTAIN_LION_OR_LATER(self.profile):
            yield self._dtb_hits_idlepml4
        else:
            yield self._dtb_hits_legacy

        if self.profile.metadata("arch") == "AMD64":
            yield self._dtb_hits_kernel_pmap

    def dtb_hits(self):
        for method in self._dtb_methods():
            for dtb_hit in method():
                yield dtb_hit

    def VerifyHit(self, hit):
        address_space = self.CreateAS(hit)

        if address_space:
            address = self.profile.get_constant(
                "_version", is_address=True)
            if not address_space.is_valid_address(address):
                return

            if address_space.read(address, 13) != "Darwin Kernel":
                return

            return address_space

    def render(self, renderer):
        renderer.table_header([("DTB", "dtb", "[addrpad]"),
                               ("Verified", "verified", "8"),
                               ("Source", "method", "15")])
        for method in self._dtb_methods():
            for dtb_hit in method():
                renderer.table_row(
                    dtb_hit,
                    self.VerifyHit(dtb_hit) is not None,
                    method.__name__)


class DarwinPlugin(DarwinKASLRMixin,
                   plugin.KernelASMixin,
                   AbstractDarwinCommandPlugin):
    """Plugin which requires the kernel Address space to be loaded."""
    __abstract = True


class DarwinProcessFilter(DarwinPlugin):
    """A class for filtering processes."""

    __abstract = True

    @classmethod
    def args(cls, parser):
        super(DarwinProcessFilter, cls).args(parser)
        parser.add_argument("--pid",
                            type="ArrayIntParser",
                            help="One or more pids of processes to select.")

        parser.add_argument("--proc_regex", default=None,
                            help="A regex to select a process by name.")

        parser.add_argument("--phys_proc",
                            type="ArrayIntParser",
                            help="Physical addresses of proc structs.")

        parser.add_argument("--proc", type="ArrayIntParser",
                            help="Kernel addresses of proc structs.")

        parser.add_argument("--first", type="IntParser",
                            help="Kernel addresses of first proc to start "
                            "following.")

        parser.add_argument(
            "--method", choices=list(cls.METHODS), nargs="+",
            help="Method to list processes (Default uses all methods).")

    def __init__(self, pid=None, proc_regex=None, phys_proc=None, proc=None,
                 first=None, method=None, **kwargs):
        """Filters processes by parameters.

        Args:
           phys_proc_struct: One or more proc structs or offsets defined in
              the physical AS.

           pids: A list of pids.
           pid: A single pid.
        """
        super(DarwinProcessFilter, self).__init__(**kwargs)

        # Per-method cache of procs discovered.
        self.cache = {}

        self.methods = method or self.METHODS

        if isinstance(phys_proc, (int, long)):
            phys_proc = [phys_proc]
        elif phys_proc is None:
            phys_proc = []

        if isinstance(proc, (int, long)):
            proc = [proc]
        elif isinstance(proc, obj.Struct):
            proc = [proc.obj_offset]
        elif proc is None:
            proc = []

        self.phys_proc = phys_proc
        self.proc = proc

        pids = []
        if isinstance(pid, list):
            pids.extend(pid)

        elif isinstance(pid, (int, long)):
            pids.append(pid)

        if self.session.pid and not pid:
            pids.append(self.session.pid)

        self.pids = pids
        self.proc_regex_text = proc_regex
        if isinstance(proc_regex, basestring):
            proc_regex = re.compile(proc_regex, re.I)

        self.proc_regex = proc_regex

        # Without a specified proc head, we use the proclist from _allproc
        # constant.
        if not first:
            first = self.profile.get_constant_object(
                "_allproc", target="proclist").lh_first

        self.first = first

        # Sometimes its important to know if any filtering is specified at all.
        self.filtering_requested = (self.pids or self.proc_regex or
                                    self.phys_proc or self.proc)

    def list_using_dead_procs(self):
        """List deallocated proc structs using the zone allocator."""
        # Find the proc zone from the allocator.
        proc_zone = self.session.manager.find_first(
            "AllocationZone/name is 'proc'")["Struct/base"]

        # Walk over the free list and get all the proc objects.
        obj_list = proc_zone.free_elements.walk_list("next")
        result = []
        for object in obj_list:
            proc = object.cast("proc")

            # Validate the proc. Real procs have a non zero argc.
            if proc.p_argc > 0:
                result.append(proc)

        return result

    def list_using_allproc(self):
        """List all processes by following the _allproc list head."""
        result = set(self.first.p_list)
        return result

    def list_using_tasks(self):
        """List processes using the processor tasks queue.

        See
        /osfmk/kern/processor.c (processor_set_things)
        """
        seen = set()

        tasks = self.profile.get_constant_object(
            "_tasks",
            target="queue_entry",
            vm=self.kernel_address_space)

        for task in tasks.list_of_type("task", "tasks"):
            proc = task.bsd_info.deref()
            if proc:
                seen.add(proc)

        return seen

    def list_using_pgrp_hash(self):
        """Process groups are organized in a hash chain.

        xnu-1699.26.8/bsd/sys/proc_internal.h
        """
        seen = set()

        # Note that _pgrphash is initialized through:

        # xnu-1699.26.8/bsd/kern/kern_proc.c:195
        # hashinit(int elements, int type, u_long *hashmask)

        # /xnu-1699.26.8/bsd/kern/kern_subr.c: 327
        # hashinit(int elements, int type, u_long *hashmask) {
        #    ...
        # *hashmask = hashsize - 1;

        # Hence the value in _pgrphash is one less than the size of the hash
        # table.
        pgr_hash_table = self.profile.get_constant_object(
            "_pgrphashtbl",
            target="Pointer",
            target_args=dict(
                target="Array",
                target_args=dict(
                    target="pgrphashhead",
                    count=self.profile.get_constant_object(
                        "_pgrphash", "unsigned long") + 1
                )
            )
        )

        for slot in pgr_hash_table.deref():
            for pgrp in slot.lh_first.walk_list("pg_hash.le_next"):
                for proc in pgrp.pg_members.lh_first.walk_list(
                        "p_pglist.le_next"):
                    seen.add(proc)

        return seen

    def list_using_pid_hash(self):
        """Lists processes using pid hash tables.

        xnu-1699.26.8/bsd/kern/kern_proc.c:834:
        pfind_locked(pid_t pid)
        """
        seen = set()

        # Note that _pidhash is initialized through:

        # xnu-1699.26.8/bsd/kern/kern_proc.c:194
        # pidhashtbl = hashinit(maxproc / 4, M_PROC, &pidhash);

        # /xnu-1699.26.8/bsd/kern/kern_subr.c: 327
        # hashinit(int elements, int type, u_long *hashmask) {
        #    ...
        # *hashmask = hashsize - 1;

        # Hence the value in pidhash is one less than the size of the hash
        # table.
        pid_hash_table = self.profile.get_constant_object(
            "_pidhashtbl",
            target="Pointer",
            target_args=dict(
                target="Array",
                target_args=dict(
                    target="pidhashhead",
                    count=self.profile.get_constant_object(
                        "_pidhash", "unsigned long") + 1
                )
            )
        )

        for plist in pid_hash_table.deref():
            for proc in plist.lh_first.walk_list("p_hash.le_next"):
                if proc:
                    seen.add(proc)

        return seen

    def list_procs(self, sort=True):
        """Uses a few methods to list the procs."""
        seen = set()

        for method, handler in self.METHODS.iteritems():
            if method not in self.methods:
                continue

            procs = self.cache.setdefault(method, handler(self))
            self.session.logging.debug("Listed %d processes using %s",
                                       len(procs), method)

            seen.update(procs)

        if sort:
            return sorted(seen, key=lambda proc: proc.p_pid)

        return seen

    def filter_processes(self):
        """Filters proc list using phys_proc and pids lists."""
        # No filtering required:
        if not self.filtering_requested:
            for proc in self.list_procs():
                yield proc
        else:
            # We need to filter by phys_proc
            for offset in self.phys_proc:
                yield self.virtual_process_from_physical_offset(offset)

            for offset in self.proc:
                yield self.profile.proc(vm=self.kernel_address_space,
                                        offset=int(offset))

            # We need to filter by pids
            for proc in self.list_procs():
                if int(proc.p_pid) in self.pids:
                    yield proc

                elif self.proc_regex and self.proc_regex.match(
                        utils.SmartUnicode(proc.p_comm)):
                    yield proc

    def virtual_process_from_physical_offset(self, physical_offset):
        """Tries to return an proc in virtual space from a physical offset.

        We do this by reflecting off the list elements.

        Args:
           physical_offset: The physcial offset of the process.

        Returns:
           an _PROC object or a NoneObject on failure.
        """
        physical_proc = self.profile.eprocess(offset=int(physical_offset),
                                              vm=self.kernel_address_space.base)

        # We cast our list entry in the kernel AS by following Flink into the
        # kernel AS and then the Blink. Note the address space switch upon
        # dereferencing the pointer.
        our_list_entry = physical_proc.procs.next.dereference(
            vm=self.kernel_address_space).prev.dereference()

        # Now we get the proc_struct object from the list entry.
        return our_list_entry.dereference_as("proc_struct", "procs")

    METHODS = {
        "allproc": list_using_allproc,
        "deadprocs": list_using_dead_procs,
        "tasks": list_using_tasks,
        "pgrphash": list_using_pgrp_hash,
        "pidhash": list_using_pid_hash,
    }


class KernelAddressCheckerMixIn(object):
    """A plugin mixin which does kernel address checks."""

    def __init__(self, **kwargs):
        super(KernelAddressCheckerMixIn, self).__init__(**kwargs)

        # We use the module plugin to help us local addresses inside kernel
        # modules.
        self.module_plugin = self.session.plugins.lsmod(session=self.session)



# Rekall Memory Forensics
# Copyright (C) 2008-2011 Volatile Systems
# Copyright (C) 2011 Jamie Levy (Gleeda) <jamie.levy@gmail.com>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Additional Authors:
# Michael Cohen <scudette@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
@author:       Jamie Levy (gleeda)
@license:      GNU General Public License 2.0 or later
@contact:      jamie.levy@gmail.com
@organization: Volatile Systems
"""
from rekall import plugin
from rekall import obj
from rekall import scan
from rekall import utils
from rekall.plugins.windows.registry import registry
from rekall.plugins.windows.registry import getsids

import ntpath


# for more information on Event Log structures see WFA 2E pg 260-263 by Harlan
# Carvey
evt_log_types = {
    'EVTLogHeader' : [0x30, {
        'HeaderSize' : [0x0, ['unsigned int']],
        'Magic' : [0x4, ['String', dict(length=4)]],  # LfLe

        # Offset of oldest record.
        'OffsetOldest' : [0x10, ['int']],

        # Offset of next record to be written.
        'OffsetNextToWrite' : [0x14, ['int']],
        'NextID' : [0x18, ['int']],  # Next event record ID.
        'OldestID' : [0x1c, ['int']], # Oldest event record ID.

        # Maximum size of event record (from registry).
        'MaxSize' : [0x20, ['int']],

        # Retention time of records (from registry).
        'RetentionTime' : [0x28, ['int']],

        # Size of the record (repeat of DWORD at offset 0).
        'RecordSize' : [0x2c, ['int']],
        }],

    'EVTRecordStruct' : [0x38, {
        'RecordLength' : [0x0, ['int']],
        'Magic' : [0x4, ['String', dict(length=4)]],  # LfLe
        'RecordNumber' : [0x8, ['int']],

        'TimeGenerated' : [0xc, ['UnixTimeStamp']],
        'TimeWritten' : [0x10, ['UnixTimeStamp']],

        # Specific to event source and uniquely identifies the event.
        'EventID' : [0x14, ['unsigned short']],
        'EventType' : [0x18, ['Enumeration', dict(
            target='unsigned short',
            choices={0x01: "Error",
                     0x02: "Warning",
                     0x04: "Info",
                     0x08: "Success",
                     0x10: "Failure"})]],

        # Number of description strings in event message.
        'NumStrings' : [0x1a, ['unsigned short']],
        'EventCategory' : [0x1c, ['unsigned short']],
        'ReservedFlags' : [0x1e, ['unsigned short']],
        'ClosingRecordNum' : [0x20, ['int']],

        # Offset w/in record of description strings.
        'StringOffset' : [0x24, ['int']],

        # Length of SID: if 0 no SID is present.
        'SidLength' : [0x28, ['int']],

        # Offset w/in record to start of SID (if present).
        'SidOffset' : [0x2c, ['int']],

        # Length of binary data of record.
        'DataLength' : [0x30, ['int']],

        # Offset of data w/in record.
        'DataOffset' : [0x34, ['int']],

        'Source': [0x38, ['UnicodeString', dict(
            length=lambda x: x.RecordLength)]],

        # The computer name is right after the Source.
        'Computer': [lambda x: x.Source.obj_offset + x.Source.obj_size,
                     ['UnicodeString', dict(
                         length=lambda x: x.RecordLength)]],

        'Sid': [lambda x: x.obj_offset + x.SidOffset.v(), ['_SID']],

        'Data':[lambda x: x.obj_offset + x.StringOffset.v(), [
            "ListArray", dict(
                target="UnicodeString",
                target_args=dict(encoding="utf16"),
                maximum_size=lambda x: x.RecordLength,
                count=lambda x: x.NumStrings)]],
        }],

    "_SID": [None, {
        "IdentifierAuthority": [None, ["Enumeration", dict(
            choices={
                "\x00\x00\x00\x00\x00\x00": "Null Authority",
                "\x00\x00\x00\x00\x00\x01": "World Authority",
                "\x00\x00\x00\x00\x00\x02": "Local Authority",
                "\x00\x00\x00\x00\x00\x03": "Creator Authority",
                "\x00\x00\x00\x00\x00\x04": "NonUnique Authority",
                "\x00\x00\x00\x00\x00\x05": "NT Authority",
                },
            target="String",
            target_args=dict(length=6, term=None)
            )]],
        "NumericIdentifier": [0x4, ["unsigned be int"]],
        "SubAuthority": [None, ["Array", dict(
            target="unsigned long",
            count=lambda x: x.SubAuthorityCount)]],
        }],
    }



class _SID(obj.Struct):
    """A Pretty printing implementation of sids.

    Reference:
    http://www.sekchek.com/downloads/white-papers/windows-about-sids.pdf
    """
    def __unicode__(self):
        """Format the Sid using SDDL Notation."""
        components = [self.Revision, self.NumericIdentifier]
        components.extend(self.SubAuthority)

        result = u"S-" + u"-".join([str(x) for x in components])

        # Try to resolve a friendly name from the cache in the context.
        friendly_name = self.obj_context.get("sid_cache", {}).get(result)
        if friendly_name:
            result = u"%s (%s)" % (result, friendly_name)

        return result


class EVTObjectTypes(obj.ProfileModification):
    """An implementation for parsing event logs."""

    @classmethod
    def modify(cls, profile):
        profile.add_overlay(evt_log_types)
        profile.add_classes(dict(_SID=_SID))


class EVTScanner(scan.BaseScanner):
    checks = [('StringCheck', dict(needle="LfLe"))]

    def scan(self, offset, maxlen=None, context=None):
        for hit in super(EVTScanner, self).scan(offset, maxlen=maxlen):
            event_offset = hit - self.profile.get_obj_offset(
                "EVTRecordStruct", "Magic")

            event = self.profile.EVTRecordStruct(
                offset=event_offset, vm=self.address_space, context=context)

            # Eliminate crazy events (between 2001 and 2017):
            if (1500000000 > event.TimeGenerated > 1000000000 and
                    1500000000 > event.TimeWritten > 1000000000):
                yield event


class EvtLogs(plugin.VerbosityMixIn, registry.RegistryPlugin):
    """Extract Windows Event Logs (XP/2003 only)"""

    __name = "evtlogs"

    @classmethod
    def is_active(cls, config):
        """Only active for windows XP."""
        return (super(EvtLogs, cls).is_active(config) and
                config.profile.metadata("major") == 5)

    def __init__(self, **kwargs):
        super(EvtLogs, self).__init__(**kwargs)
        self.profile = EVTObjectTypes(self.profile)
        self.context = dict(sid_cache={})

    def FindEVTFiles(self):
        """Search for event log files in memory.

        We search for processes called 'services.exe' with a vad to and open
        file ending with '.evt'.
        """
        ps_plugin = self.get_plugin("pslist", proc_regex="services.exe")

        for task in ps_plugin.filter_processes():
            for vad in task.RealVadRoot.traverse():
                try:
                    filename = vad.ControlArea.FilePointer.FileName
                    if utils.SmartUnicode(filename).lower().endswith(".evt"):
                        yield task, vad
                except AttributeError:
                    pass

    def ScanEvents(self, vad, address_space):
        scanner = EVTScanner(profile=self.profile, address_space=address_space,
                             session=self.session)
        for event in scanner.scan(offset=vad.Start, maxlen=vad.Length,
                                  context=self.context):
            yield event

    def PrecacheSids(self):
        """Search for known sids that we can cache."""
        sid_cache = self.context["sid_cache"]
        sid_cache.update(getsids.well_known_sids)

        # Search for all known user sids.
        for hive_offset in self.hive_offsets:
            hive_address_space = registry.HiveAddressSpace(
                base=self.kernel_address_space,
                hive_addr=hive_offset, profile=self.profile)

            reg = registry.Registry(
                profile=self.profile, address_space=hive_address_space)

            # We get the user names according to the name of the diretory where
            # their profile is. This is not very accurate - should we check the
            # SAM instead?
            profiles = reg.open_key(
                'Microsoft\\Windows NT\\CurrentVersion\\ProfileList')

            for profile in profiles.subkeys():
                path = profile.open_value("ProfileImagePath").DecodedData
                if path:
                    sid_cache[utils.SmartUnicode(profile.Name)] = (
                        utils.SmartUnicode(ntpath.basename(path)))

        # Search for all service sids.
        getservicesids = self.get_plugin("getservicesids")
        for sid, service_name in getservicesids.get_service_sids():
            sid_cache[sid] = "(Service: %s)" % service_name

    def render(self, renderer):
        if self.verbosity > 5:
            self.PrecacheSids()

        renderer.table_header([("TimeWritten", "timestamp", ""),
                               ("Filename", "filename", ""),
                               ("Computer", "computer", ""),
                               ("Sid", "sid", ""),
                               ("Source", "source", ""),
                               ("Event Id", "event_id", ""),
                               ("Event Type", "event_type", ""),
                               ("Message", "message", "")])

        for task, vad in self.FindEVTFiles():
            filename = ntpath.basename(
                utils.SmartUnicode(vad.ControlArea.FilePointer.FileName))

            for event in self.ScanEvents(vad, task.get_process_address_space()):
                args = ";".join(
                    repr(utils.SmartStr(x)) for x in event.Data)

                renderer.table_row(
                    event.TimeWritten,
                    filename,
                    event.Computer,
                    event.Sid,
                    event.Source,
                    event.EventID,
                    event.EventType,
                    args)



# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""Various plugins that make entities testable and easier to profile."""

__author__ = "Adam Sindelar <adamsh@google.com>"

import itertools

from rekall import plugin
from rekall import testlib

from rekall.entities import entity as entity_module
from rekall.entities import component as entity_component

from efilter import expression as expr
from efilter import query as entity_query


class TestEntityFind(testlib.SortedComparison):
    PARAMETERS = dict(
        commandline="find \"Process/command =~ '%(process)s'\"",
        process="lsass")


class TestEntityAnalyze(testlib.SortedComparison):
    PARAMETERS = dict(commandline="analyze \"Process/command =~ 'lsass'\"")


class EntityAnalyze(plugin.ProfileCommand):
    __name = "analyze"

    COLLECTOR_COSTS = ["none", "cheap", "normal", "high", "INSANE!"]

    @classmethod
    def args(cls, parser):
        super(EntityAnalyze, cls).args(parser)
        parser.add_argument("query", positional=True,
                            help="The filter query to use")

    def __init__(self, query=None, **kwargs):
        super(EntityAnalyze, self).__init__(**kwargs)
        self.query = entity_query.Query(query)
        self.analysis = self.session.entities.analyze(self.query)

    def render_components(self, renderer):
        renderer.section("Expected components in result set:", width=140)
        renderer.table_header([
            dict(name="Guaranteed", width=20),
            dict(name="Component", width=20)])

        for component in self.analysis["guaranteed_components"]:
            renderer.table_row(True, component)

        for component in self.analysis["possible_components"]:
            renderer.table_row(False, component)

    def render_dependencies(self, renderer):
        renderer.section("Expected requirements for matching objects:",
                         width=140)
        renderer.table_header([
            dict(name="Dependency", cname="dependency", width=40,
                 type="Dependency"),
            dict(name="Explanation", cname="explanation", width=100,
                 type="Query")])

        for dependency in itertools.chain(self.analysis["dependencies"],
                                          self.analysis["exclusions"]):
            renderer.table_row(dependency, self.query,
                               query_highlight=dependency.expression)

    def render_indexing(self, renderer):
        renderer.section("Suggested indexing (created automatically).")
        renderer.table_header([
            dict(name="Indexed attribute", width=40)])
        for lookup in self.analysis["lookups"]:
            renderer.table_row(lookup)

    def render_collectors(self, renderer):
        manager = self.session.entities
        renderer.section("Dependencies on collectors:", width=140)
        renderer.table_header([
            dict(name="Collector", cname="collector", width=40),
            dict(name="Cost to run", cname="cost", width=20),
            dict(name="Needed because of", cname="reason", width=80)])

        collectors = {}
        queue = []
        for collector in self.analysis["collectors"]:
            queue.append(collector)
            collectors[collector] = ["the query"]

        while queue:
            collector = queue.pop()
            for query in collector.collect_queries.itervalues():
                for additional in manager.analyze(query)["collectors"]:
                    if additional in collectors:
                        collectors[additional].append(collector.name)
                        continue
                    else:
                        collectors[additional] = [collector.name]
                        queue.append(additional)

        for collector, reasons in collectors.iteritems():
            renderer.table_row(collector.name,
                               self.COLLECTOR_COSTS[collector.run_cost],
                               ", ".join(reasons))

    def _render_node(self, node, renderer, depth=0):
        renderer.table_row(
            type(node).__name__,
            self.query,
            depth=depth, query_highlight=node)

        for child in node.children:
            if not isinstance(child, expr.Expression):
                renderer.table_row(
                    "leaf (%s):" % type(child).__name__,
                    "",
                    depth=depth + 1)
                continue

            self._render_node(node=child, renderer=renderer, depth=depth + 1)

    def render_tree(self, renderer):
        renderer.section("Query analysis:", width=140)
        renderer.table_header([
            dict(name="Expression", cname="expression", type="TreeNode",
                 max_depth=15, width=40),
            dict(name="Location in query", cname="source", type="Query",
                 width=100)])

        self._render_node(self.query.root, renderer)

    def render(self, renderer):
        self.render_tree(renderer)
        self.render_dependencies(renderer)
        self.render_components(renderer)
        self.render_collectors(renderer)
        self.render_indexing(renderer)


class EntityFind(plugin.ProfileCommand):
    """Runs the query and displays results. Base class for saved searches.

    This class is designed to be subclasses by saved searches. Subclasses may
    override the defaults for the following properties:

    query, search: The query that'll be run.
    (--)display_filter: Another query that'll be used to filter out unwanted
                        results.
    (--)columns: A list of attributes to render in the output. Format as
                 "Component/attribute", or use the column DSL defined below.
    (--)sort: A list of columns to sort by. Sort is currently ASC. Same format
              as above. Mutually exclusive with --stream_results.
    (--)width: Width of the rendered table.
    (--)stream_results: If on, will render results as soon as they're available.
                        This typically means the results will be incomplete. It
                        is probably pointless to use this in combination with
                        'sort' but no one will stop you. Mutually exclusive
                        with --sort.

    Further arguments that can be supplied at runtime:

    --explain: If set, an analysis of the query will be rendered and each
               row in results will include a highlight of the part of the query
               that matched it (obviously a heuristic, your mileage may vary).

    Column definitions:
    ===================

    Columns can be given as a string or a dict. A string is interpreted as
    shorthand for dict(attribute=STRING).

    A column may specify any of the following keys in the dict:
    - name: The name of the table column as displayed to the user. If not given
            will be derived from the attribute name.
    - attribute: The name of the attribute to be selected from each entity.
    - fn: A callable that is passed an entity and produces a value of the
          column. You can only provide this or attribute but not both.
    - width: The width of the column (number of characters). If not given, will
             be derived from the attribute's default width (if that is given).
    """

    __name = "find"
    description = None

    search = None
    display_filter = None
    columns = ()
    _entity_columns = None
    _table_columns = None
    sort = ()
    width = 120
    stream_results = False
    complete_results = True
    syntax = "slashy"

    query = None

    @classmethod
    def args(cls, parser):
        super(EntityFind, cls).args(parser)
        parser.add_argument("query", required=False,
                            help="The filter query to use.")
        parser.add_argument("--columns", default=None, nargs="+", type="str")
        parser.add_argument("--sort", default=None, nargs="+", type="str")
        parser.add_argument("--explain", type="Boolean", default=False,
                            help="Show which part of the query matched.")
        parser.add_argument("--width", type="int", default=None)
        parser.add_argument("--filter", type="str")
        parser.add_argument("--complete_results", type="Boolean", default=True)
        parser.add_argument("--stream_results", type="Boolean", default=False)

    def __init__(self, query=None, explain=None, columns=None, sort=None,
                 width=None, filter=None, stream_results=False,
                 complete_results=True, syntax="slashy", **kwargs):
        super(EntityFind, self).__init__(**kwargs)
        if query:
            self.query = entity_query.Query(query)
        else:
            self.query = entity_query.Query(self.search)

        if columns is not None:
            self.columns = columns

        if sort is not None:
            self.sort = sort

        if syntax is not None:
            self.syntax = syntax

        if width is not None:
            self.width = width

        if filter is not None:
            self.display_filter = entity_query.Query(filter)

        if stream_results is not None:
            if sort:
                raise ValueError(
                    "Can't stream results and sort at the same time.")
            self.stream_results = stream_results

        if complete_results is not None:
            self.complete_results = complete_results

        self.explain = explain

    def _build_sort_func(self):
        if not self.sort:
            return
        sort_keys = tuple(self.sort)
        sort_types = [entity_module.Entity.reflect_attribute(a).typedesc
                      for a in sort_keys]

        # Entity is in the first column. Rows are tuples of (object, options).
        def _sort_func(entity):
            result = []
            for idx, key in enumerate(sort_keys):
                result.append(sort_types[idx].sortkey(entity[key]))
            return result

        return _sort_func

    def render_entity(self, renderer, entity):
        if self.display_filter and not self.display_filter.run_engine(
                "matcher", bindings=entity):
            return

        opts = {}
        values = []
        for column in self.entity_columns:
            attribute = column.get("attribute")
            fn = column.get("fn")
            if attribute:
                if fn:
                    raise ValueError(
                        "Column spec %r provides both a fn lambda and an "
                        "attribute. Pick one or the other." % column)
                values.append(entity[attribute])
            elif fn:
                values.append(fn(entity))
            else:
                raise ValueError(
                    "Column spec %r doesn't specify attribute name or lambda."
                    % column)

        if self.explain:
            match = self.query.run_engine("matcher",
                                          bindings=entity,
                                          match_backtrace=True)
            values.append(self.query)
            opts["query_highlight"] = match.matched_expression

        renderer.table_row(*values, **opts)

    @property
    def entity_columns(self):
        if self._entity_columns:
            return self._entity_columns

        if not self.columns:
            self.columns = []
            for component in self.session.entities.analyze(self.query).get(
                    "guaranteed_components"):
                component_cls = entity_module.Entity.reflect_component(
                    component)
                for attribute in component_cls.component_attributes.values():
                    if attribute.hidden:
                        continue

                    self.columns.append(attribute.path)

            self.columns.sort()

        self._entity_columns = []
        for column in self.columns:
            if isinstance(column, basestring):
                self._entity_columns.append(dict(attribute=column))
            else:
                self._entity_columns.append(column)

        return self._entity_columns

    @property
    def table_columns(self):
        if self._table_columns:
            return self._table_columns

        self._table_columns = []
        for column in self.entity_columns:
            result = dict(name="(untitled column)", style=None, width=None)
            name = column.get("name")
            attribute = column.get("attribute")

            if attribute:
                field = entity_module.Entity.reflect_attribute(attribute)

                # TODO(scudette): This seems to confuse the renderer system for
                # whatever reason. It wants to guess the type all by itself.
                # result["type"] = field.typedesc.type_name

                result["name"] = field.name
                result["width"] = field.width
                result["style"] = field.style

            if name:
                result["name"] = name

            result["width"] = column.get("width", result["width"])
            result["style"] = column.get("style", result["style"])

            self._table_columns.append(result)

        if self.explain:
            self._table_columns.append(dict(name="Matched query", type="Query",
                                            width=60))

        return self._table_columns

    def render(self, renderer):
        if self.explain:
            self.session.RunPlugin("analyze", self.query)
            renderer.section("Results:", width=self.width)

        renderer.table_header(self.table_columns)

        self.session.report_progress(
            "Running query %(query)s %(spinner)s", query=self.query)

        if self.stream_results:
            def _handler(entity):
                self.render_entity(renderer, entity)

            self.session.entities.stream(self.query, _handler)
        else:
            rows = self.session.entities.find(self.query,
                                              complete=self.complete_results,
                                              syntax=self.syntax)
            if self.sort:
                rows = sorted(rows, key=self._build_sort_func())
            for entity in rows:
                self.render_entity(renderer, entity)


class FindBatch(plugin.ProfileCommand):
    """Runs several plugins in order. Subclass to set the batch."""

    __abstract = True
    batch = []

    @classmethod
    def args(cls, parser):
        super(FindBatch, cls).args(parser)
        parser.add_argument("--sort", default=None, nargs="+", type="str")
        parser.add_argument("--explain", type="Boolean", default=False,
                            help="Show which part of the query matched.")
        parser.add_argument("--filter", type="str")

    def __init__(self, filter=None, sort=None, explain=None, **kwargs):
        super(FindBatch, self).__init__(**kwargs)
        self.display_filter = filter
        self.sort = sort
        self.explain = explain

    def render(self, renderer):
        for command in self.batch:
            plugin_obj = getattr(self.session.plugins, command)()
            description = getattr(plugin_obj, "description")
            if description:
                renderer.section(description, width=120)

            self.session.RunPlugin(
                command, filter=self.display_filter, sort=self.sort,
                explain=self.explain)


class EntityDescribe(plugin.ProfileCommand):
    """Analyzes the query and prints a lot of mostly useless information."""
    __name = "describe"

    @classmethod
    def args(cls, parser):
        super(EntityDescribe, cls).args(parser)
        parser.add_positional_arg("component")

    def __init__(self, component=None, **kwargs):
        super(EntityDescribe, self).__init__(**kwargs)
        self.component = component

    def render_component(self, renderer, component_cls):
        renderer.section(
            "%s: %s" % (component_cls.component_name,
                        component_cls.component_docstring),
            width=140)
        renderer.table_header([
            dict(name="Field", cname="field", width=16),
            dict(name="Type", cname="type", width=20),
            dict(name="Description", cname="description", width=60),
            dict(name="Show by default", cname="visible", width=16),
            dict(name="Notes", cname="notes", width=30)])

        for attribute in sorted(
                component_cls.component_attributes.itervalues(),
                key=lambda attr: attr.name):
            alias = getattr(attribute, "alias", None)
            renderer.table_row(attribute.name,
                               "%s (%s)" % (attribute.typedesc.type_name,
                                            type(attribute).__name__),
                               attribute.docstring,
                               not attribute.hidden,
                               "Aliased to %s" % alias if alias else None)

    def render(self, renderer):
        if self.component:
            return self.render_component(
                renderer,
                entity_component.Component.classes[self.component])

        for component_cls in sorted(
                entity_component.Component.classes.itervalues(),
                key=lambda comp: comp.component_name):
            self.render_component(renderer, component_cls)



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
Collectors for files, handles, sockets and similar.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

from rekall.entities import collector
from rekall.entities import definitions

from rekall.plugins.collectors.darwin import common


class DarwinHandleCollector(common.DarwinEntityCollector):
    """Collects handles from fileprocs (like OS X lsof is implemented)."""

    _name = "handles"
    outputs = ["Handle",
               "Struct/type=fileproc",
               "Struct/type=vnode",
               "Struct/type=socket"]
    collect_args = dict(processes="has component Process")
    filter_input = True

    run_cost = collector.CostEnum.HighCost

    def collect(self, hint, processes):
        manager = self.manager
        for process in processes:
            proc = process["Struct/base"]

            for fd, fileproc, flags in proc.get_open_files():
                fg_data = fileproc.autocast_fg_data()

                # The above can return None if the data in memory is invalid.
                # There's nothing we can do about that, other than rely on
                # collector redundancy. Skip.
                if not fg_data:
                    continue

                # In addition to yielding the handle, we will also yield the
                # resource it's pointing to, because other collectors rely on
                # memory objects already being out there when they parse them
                # for resource (File/Socket/etc.) specific information.
                resource_identity = manager.identify({
                    "Struct/base": fg_data})
                handle_identity = manager.identify({
                    "Struct/base": fileproc})

                yield [
                    resource_identity,
                    definitions.Struct(
                        base=fg_data,
                        type=fg_data.obj_type)]

                yield [
                    handle_identity,
                    definitions.Handle(
                        process=process.identity,
                        fd=fd,
                        flags=flags,
                        resource=resource_identity),
                    definitions.Struct(
                        base=fileproc,
                        type="fileproc")]


class DarwinFileCollector(common.DarwinEntityCollector):
    """Collects files based on vnodes."""

    outputs = ["File", "Permissions", "Timestamps", "Named"]
    _name = "files"
    collect_args = dict(vnodes="Struct/type is 'vnode'")
    filter_input = True

    def collect(self, hint, vnodes):
        manager = self.manager
        for entity in vnodes:
            vnode = entity["Struct/base"]
            path = vnode.full_path

            components = [entity.identity,
                          definitions.File(
                              path=path),
                          definitions.Named(
                              name=path,
                              kind="File")]

            # Parse HFS-specific metadata. We could look at the mountpoint and
            # see if the filesystem is actually HFS, but it turns out that
            # cnodes are also used for stuff like the dev filesystem, so let's
            # just try and see if there's one that looks valid and go with it.
            cnode = vnode.v_data.dereference_as("cnode")
            if cnode.c_rwlock == cnode:
                cattr = vnode.v_data.dereference_as("cnode").c_attr

                # HFS+ stores timestamps as UTC.
                components.append(definitions.Timestamps(
                    created_at=cattr.ca_ctime.as_datetime(),
                    modified_at=cattr.ca_mtime.as_datetime(),
                    accessed_at=cattr.ca_atime.as_datetime(),
                    backup_at=cattr.ca_btime.as_datetime()))

            posix_uid = vnode.v_cred.cr_posix.cr_ruid
            if posix_uid and posix_uid != 0:
                components.append(definitions.Permissions(
                    owner=manager.identify({
                        "User/uid": posix_uid})))

            yield components



from rekall import addrspace
from rekall import obj
from rekall import testlib
from rekall import session


class CustomRunsAddressSpace(addrspace.RunBasedAddressSpace):
    def __init__(self, runs=None, data=None, **kwargs):
        super(CustomRunsAddressSpace, self).__init__(**kwargs)
        self.base = addrspace.BufferAddressSpace(data=data,
                                                 session=self.session)
        for i in runs:
            self.runs.insert(i)


class RunBasedTest(testlib.RekallBaseUnitTestCase):
    """Test the RunBasedAddressSpace implementation."""

    def setUp(self):
        self.session = session.Session()
        self.contiguous_as = CustomRunsAddressSpace(session=self.session,
            runs = [(1000, 0, 1), (1001, 1, 9)],
            data="0123456789")
        self.discontiguous_as = CustomRunsAddressSpace(session=self.session,
            runs=[(1000, 0, 1), (1020, 1, 9)],
            data="0123456789")

    def testDiscontiguousRunsRead(self):
        # Read from an address without data
        self.assertEqual(self.discontiguous_as.read(0, 20),
                         "\x00" * 20)
        # Read spanning two runs
        self.assertEqual(self.discontiguous_as.read(1000, 30),
                         "0" + "\x00"*19 + "123456789" +  "\x00")
        # Read in the middle of a run
        self.assertEqual(self.discontiguous_as.read(1025, 10),
                         "6789" + "\x00" * 6)
        # Read past the end
        self.assertEqual(self.discontiguous_as.read(2000, 10),
                         "\x00" * 10)

    def testContiguousRunsRead(self):
        # Read from an address without data
        self.assertEqual(self.contiguous_as.read(0, 20),
                         "\x00" * 20)
        # Read spanning two runs
        self.assertEqual(self.contiguous_as.read(1000, 30),
                         "0123456789" + "\x00"*20)
        # Read in the middle of a run
        self.assertEqual(self.contiguous_as.read(1005, 10),
                         "56789" + "\x00" * 5)
        # Read past the end
        self.assertEqual(self.contiguous_as.read(2000, 10),
                         "\x00" * 10)

if __name__ == "__main__":
    unittest.main()



# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""This module implements renderers specific to Linux structures."""

from rekall.ui import json_renderer
from rekall.ui import text


class kuid_t_TextObjectRenderer(text.TextObjectRenderer):
    renders_type = "kuid_t"
    renderers = ["TextRenderer", "TestRenderer", "WideTextRenderer"]

    def render_row(self, target, **_):
        return text.Cell(unicode(target))

class kgid_t_TextObjectRenderer(kuid_t_TextObjectRenderer):
    renders_type = "kgid_t"


class kuid_t_JsonObjectRenderer(json_renderer.JsonObjectRenderer):
    renders_type = ["kuid_t", "kgid_t"]
    renderers = ["JsonRenderer", "DataExportRenderer"]

    def EncodeToJsonSafe(self, task, **_):
        return task.val.v()



"""Collectors specific to Windows"""


from rekall.plugins.collectors.windows import common
from rekall.plugins.collectors.windows import handles
from rekall.plugins.collectors.windows import networking
from rekall.plugins.collectors.windows import pool
from rekall.plugins.collectors.windows import processes



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
The Rekall Entity Layer.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"


from efilter import expression


class IdentityError(RuntimeError):
    pass


class Identity(object):
    """Uniquely identifies something like a process or a user.

    Instantiate with:
    - global_prefix: The universe this identity belongs to.
    - indices: Iterable of tuples of (prefix, attribute_name, value) that
        identify some entity.
    """

    global_prefix = None
    key_canary = None  # Set of attributes references in indices.
    indices = None  # Set of tuples of (global_prefix, attribute, value).
    first_index = None  # Used for hashing.
    name = None  # MAY store name of target entity.
    kind = None  # MAY store kind of target entity.

    @classmethod
    def from_dict(cls, global_prefix, identity_dict):
        indices = []
        for key, val in identity_dict.iteritems():
            if val == None:
                raise IdentityError(
                    "Identity index value for %s cannot be None." % key)

            indices.append((global_prefix, key, val))

        return cls(indices, global_prefix)

    def as_query(self):
        """A query that'll match entities with this identity."""
        union = []
        for _, keys, vals in self.indices:
            if isinstance(keys, tuple):
                intersection = []
                for idx, key in enumerate(keys):
                    intersection.append(
                        expression.Equivalence(
                            expression.Binding(key),
                            expression.Literal(vals[idx])))

                union.append(expression.Intersection(*intersection))
            else:
                union.append(expression.Equivalence(
                    expression.Binding(keys),
                    expression.Literal(vals)))

        if len(union) == 1:
            return union[0]

        return expression.Union(*union)

    def __init__(self, indices=None, global_prefix=None):
        self.key_canary = frozenset([attribute for _, attribute, _ in indices])
        self.indices = frozenset(indices)
        self.global_prefix = global_prefix

        for index in indices:
            self.first_index = index
            break

    def __hash__(self):
        """Hashing an identity has a big caveat.

        Because an identity is hashed on all of its indices, and comparisons
        between identities are done by intersection of indices, it is possible
        for two identities to have different hashes and still be equal if
        compared. This is fine in certain cases, such as with identities
        retrieved from the EntityManager (because they're already merged and
        therefore unique) but may result in unexpected behavior with identities
        made up outside of the normal collection mechanism.
        """
        return hash(self.first_index)

    def __eq__(self, other):
        if ((not isinstance(other, Identity)) or
                (self.global_prefix != other.global_prefix)):
            return False

        matching_keys = self.key_canary & other.key_canary
        if not matching_keys:
            return None  # No overlap - undecidable.

        matching_indices = self.indices & other.indices
        if not matching_indices:
            return False

        # If some keys matched but not others then that's a consistency error
        # most likely caused by faulty collector logic. By blowing up here
        # we preserve the integrity of the database, but this is a programmer
        # error and likely means that the data isn't reliable anyway.
        if len(matching_keys) != len(matching_indices):
            raise IdentityError(
                "Identity logic error! Identity %s matches %s on %d keys, "
                "but on %d values." % (
                    self, other, len(matching_keys), len(matching_indices)))

        return True

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is None:
            return None

        return not result

    def __unicode__(self):
        return repr(self)

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        indices = ", ".join([repr(x) for x in self.indices])
        if self.name or self.kind:
            return "Identity(name=%r, kind=%r, indices=[%s])" % (self.name,
                                                                 self.kind,
                                                                 indices)
        else:
            return "Identity(%s)" % indices

    def __or__(self, other):
        return self.union(other)

    def union(self, other):
        if self != other:
            raise IdentityError(
                "Attempting to merge identities %s and %s which are unequal."
                % (self, other))

        return Identity(indices=self.indices | other.indices,
                        global_prefix=self.global_prefix)



# Rekall Memory Forensics
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
Various functions for handling colors. Used mainly for visualizing output
of plugins with heatmaps (for now).

Most of the code below is made to match this color chart:
http://en.wikipedia.org/wiki/File:Xterm_256color_chart.svg

The colorspace conversions are thin wrappers around colorsys, except for
the code to handle XTerm colors, which is my own work.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

import colorsys


def ArbitraryStepFunction(value, steps):
    for step, ceiling in enumerate(steps):
        if int(value) <= ceiling:
            return step

    raise ValueError("Maximum value of %d exceeded with %d." % (steps[-1],
                                                                value))


XTERM16 = ((0x00, 0x00, 0x00), (0x80, 0x00, 0x00), (0x00, 0x80, 0x00),
           (0x80, 0x80, 0x00), (0x00, 0x00, 0x80), (0x80, 0x00, 0x80),
           (0x00, 0x80, 0x80), (0xc0, 0xc0, 0xc0), (0x80, 0x80, 0x80),
           (0xff, 0x00, 0x00), (0x00, 0xff, 0x00), (0xff, 0xff, 0x00),
           (0x00, 0x00, 0xff), (0xff, 0x00, 0xff), (0x00, 0xff, 0xff),
           (0xff, 0xff, 0xff))
"""XTerm has 16 special colors, as listed above."""


XTERM_CHANNEL_STEPS = [0, 0x5f, 0x87, 0xaf, 0xd7, 0xff]
"""XTerm color space is sparse at low luminosity."""


def ChannelStepFunction(intensity):
    return ArbitraryStepFunction(intensity, XTERM_CHANNEL_STEPS)


def GreyscaleStepFunction(intensity):
    return ArbitraryStepFunction(intensity, xrange(0x8, 0xef, 0xa))


# Color-space conversions:


def RGBToXTerm(red, green, blue):
    """Convert RGB values (0-255) to the closes XTerm color."""
    sred = ChannelStepFunction(red)
    sgreen = ChannelStepFunction(green)
    sblue = ChannelStepFunction(blue)

    # Greyscale starts at xterm 232 and has 12 shades. Black and white are part
    # of the 16-color range at the base of the spectrum.
    if sred == sgreen == sblue:
        avg = (red + green + blue) / 3
        if avg < 0x8:
            return 0
        elif avg > 0xee:
            return 15
        else:
            return 232 + GreyscaleStepFunction(avg)

    return (16  # base offset
            + ChannelStepFunction(blue)  # Blue increases in the inner loop.
            + ChannelStepFunction(green) * 6  # Green increases in the middle.
            + ChannelStepFunction(red) * 6 ** 2)  # Outer loop for red.


def XTermToRGB(xterm):
    """Convert the XTerm color (0-255) to an RGB equivalent."""
    if xterm < 16:
        return XTERM16[xterm]

    if xterm >= 232:
        # Greyscale
        value = (xterm - 231) * 0x08
        return value, value, value

    xterm -= 16  # Base of 256-color space.
    red, r = divmod(xterm, 6 ** 2)
    green, blue = divmod(r, 6)

    return (XTERM_CHANNEL_STEPS[red],
            XTERM_CHANNEL_STEPS[green],
            XTERM_CHANNEL_STEPS[blue])


def RGBToHSL(red, green, blue):
    hue, luminosity, saturation = colorsys.rgb_to_hls(
        float(red) / 0xff, float(green) / 0xff, float(blue) / 0xff)

    return hue, saturation, luminosity


def RGBToYIQ(red, green, blue):
    return colorsys.rgb_to_yiq(
        float(red) / 0xff, float(green) / 0xff, float(blue) / 0xff)


def HSLToRGB(hue, saturation, luminosity):
    red, green, blue = colorsys.hls_to_rgb(hue, luminosity, saturation)
    return int(red * 0xff), int(green * 0xff), int(blue * 0xff)


def YIQToRGB(y, i, q):
    red, green, blue = colorsys.yiq_to_rgb(y, i, q)
    return int(red * 0xff), int(green * 0xff), int(blue * 0xff)


# Text-color heuristics:


def YIQTextForBackground(y, i, q):
    """Compute the foreground color, given the background color."""
    # Y is luma, which is basically the sum of gamma-adjusted RGB channels.
    # The Y channel is intentionally weighted towards the red end of the
    # spectrum and high luminosities; I and Q are chromatic channels and carry
    # no luminosity information.
    # Perceptually, white text on red background of equivalent luminosity
    # is more readable than white text on blue or green backgrounds, hence
    # the formula below. (For greyscale this is basically identical to
    # the HSL luminosity channel, using threshold of .5, because of the gamma
    # compression's nonlinearity.)
    return (0, 0, 0) if (y * 2 - i - q) > .8 else (1, 0, 0)


def RGBTextForBackground(red, green, blue):
    """Compute the foreground color, given the background color."""
    hsl = RGBToYIQ(red, green, blue)
    text = YIQTextForBackground(*hsl)
    r, g, b = YIQToRGB(*text)
    return r, g, b


def XTermTextForBackground(xterm):
    """Compute the foreground color, given the background color."""
    rgb = XTermToRGB(xterm)
    fg = RGBTextForBackground(*rgb)
    return RGBToXTerm(*fg)


# Functions to color heatmaps:


def BlendRGB(x, y, wx=1, wy=1):
    """Blend RGB colors x and y, optionally using assigned weights wx and wy."""
    t = wx + wy
    return ((x[0] * wx + y[0] * wy) / t,
            (x[1] * wx + y[1] * wy) / t,
            (x[2] * wx + y[2] * wy) / t)


def HeatToHSL(heat, greyscale=False):
    """Given heat (0-1.0), compute the color to represent it on a heatmap.

    Arguments:
        Greyscale: If True, use luminosity instead of hue.
    """
    if greyscale:
        saturation = 0
        hue = 0
        luminosity = heat
    else:
        saturation = 1.0
        luminosity = .5
        hue = .5 - (heat * .5)

    return hue, saturation, luminosity


def HeatToRGB(heat, greyscale=False):
    return HSLToRGB(*HeatToHSL(heat, greyscale))


def HeatToXTerm(heat, greyscale=False):
    hsl = HeatToHSL(heat, greyscale)
    rgb = HSLToRGB(*hsl)
    xterm = RGBToXTerm(*rgb)

    return xterm



import logging

from rekall import addrspace
from rekall import obj

# Import and register all the plugins.
from rekall import plugins # pylint: disable=unused-import
from rekall import session
from rekall import testlib


class ProfileTest(testlib.RekallBaseUnitTestCase):
    """Test the profile implementation."""

    def setUp(self):
        self.session = session.Session()
        # Create an address space from a buffer for testing
        self.address_space = addrspace.BufferAddressSpace(
            data="hello world" * 100, session=self.session)

    def testNativeTypes(self):
        # We build a simple profile with just the native types here.
        profile = obj.Profile.classes['Profile32Bits'](session=self.session)

        # Check that simple types work
        self.assertEqual(
            profile.Object("long", offset=0, vm=self.address_space),
            0x6c6c6568)

        self.assertEqual(
            profile.Object("long long", offset=0, vm=self.address_space),
            0x6f77206f6c6c6568)

    def testBitField(self):
        # We build a simple profile with just the native types here.
        profile = obj.Profile.classes['Profile32Bits'](session=self.session)
        profile.add_types({
            'Test': [0x10, {
                'Field1': [0x00, ['BitField', dict(start_bit=0, end_bit=4)]],
                'Field2': [0x00, ['BitField', dict(start_bit=4, end_bit=8)]],
                }]})

        test = profile.Object("Test", offset=0, vm=self.address_space)

        bf1 = test.Field1
        bf2 = test.Field2

        self.assertEqual(bf1, 8)
        self.assertEqual(bf2, 6)

        # Check for overloaded numeric methods.
        self.assertEqual(bf1 + 1, 9)
        self.assertEqual(bf1 - 1, 7)
        self.assertEqual(bf1 + bf2, 8 + 6)
        self.assertEqual(bf1 < bf2, 8 < 6)
        self.assertEqual(bf1 > bf2, 8 > 6)
        self.assertEqual(bf1 & bf2, 8 & 6)
        self.assertEqual(bf1 ^ bf2, 8 ^ 6)
        self.assertEqual(bf1 + 6, 8 + bf2)
        self.assertEqual(bf1 < 6, 8 < bf2)
        self.assertEqual(bf1 > 6, 8 > bf2)
        self.assertEqual(bf1 & 6, 8 & bf2)
        self.assertEqual(bf1 ^ 6, 8 ^ bf2)

    def testPointer(self):
        # Create an address space from a buffer for testing
        address_space = addrspace.BufferAddressSpace(
            data="\x08\x00\x00\x00\x00\x00\x00\x00"
            "\x66\x55\x44\x33\x00\x00\x00\x00"
            "\x99\x88\x77\x66\x55\x44\x33\x22", session=self.session)

        vtype = {'Test': [0x10, {
            # Check simple type dereferencing
            'ptr32': [0x00, ['Pointer', dict(
                target='unsigned long'
                )]],
            'ptr64': [0x00, ['Pointer', dict(
                target='long long'
                )]],

            # Check struct dereferencing
            'next': [0x00, ['Pointer', dict(
                target='Test'
                )]],

            # A pointer to an invalid location
            'invalid': [0x08, ['Pointer', dict(
                target='long'
                )]],

            # A void pointer
            'void': [0x00, ['Pointer', dict(
                target='Void'
                )]],
            }]}

        # We build a simple profile with just the native types here.
        profile = obj.Profile.classes['Profile32Bits'](session=self.session)
        profile.add_types(vtype)

        test = profile.Object("Test", offset=0, vm=address_space)

        ptr = test.ptr32

        # Can we check the offset of members?
        self.assertEqual(profile.get_obj_offset("Test", "invalid"), 8)

        # 32 bit pointers.
        self.assertEqual(ptr.obj_size, 4)

        # The pointer itself is at location 0.
        self.assertEqual(ptr.obj_offset, 0)

        # But is pointing to location 8.
        self.assertEqual(ptr.v(), 8)
        self.assertEqual(int(ptr), 8)
        self.assertEqual(ptr, 8)

        # The long is the next 8 bytes.
        self.assertEqual(ptr.dereference(), 0x33445566)

        # Pointer comparison
        self.assertEqual(test.ptr32, test.ptr64)

        # We could do pointer arithmetic.
        ptr2 = ptr + 2

        # The new pointer is at location 8 (its 32 bits).
        self.assertEqual(ptr2.obj_offset, 8)

        # The pointer to long long is moving twice as fast
        self.assertEqual(test.ptr64 + 1, 0x33445566)
        self.assertEqual(test.ptr32 + 1, 0)

        # And its pointing to.
        self.assertEqual(ptr2.v(), 0x33445566)

        # The above makes the pointer invalid, so dereferencing it returns a 0.
        # (This is because there is no good way to validate pages except at
        # system runtime.)
        self.assertEqual(ptr2.dereference(), 0)

        # This is also invalid and will return a zero.
        self.assertEqual(test.invalid.dereference(), 0)

        # Test nonzero.
        self.assert_(test.ptr32)

        # Now dereference a struct.
        ptr3 = test.next

        # This struct starts at offset 8.
        self.assertEqual(test.next.v(), 8)

        next = ptr3.dereference()

        # We get another struct from this.
        self.assertEqual(next.obj_type, "Test")

        # This new struct's ptr32 is pointing at this address now.
        self.assertEqual(next.ptr32, 0x33445566)

        # Now test 64 bit pointers.
        profile = obj.Profile.classes['ProfileLLP64'](session=self.session)
        profile.add_types(vtype)

        test = profile.Object("Test", offset=0, vm=address_space)

        ptr = test.ptr32

        # 64 bit pointers.
        self.assertEqual(ptr.obj_size, 8)

        # The pointer itself is at location 0.
        self.assertEqual(ptr.obj_offset, 0)

        # But is pointing to location 8.
        self.assertEqual(ptr.v(), 8)

        # The long is the next 8 bytes.
        self.assertEqual(ptr.dereference(), 0x33445566)

        # We could do pointer arithmetic.
        ptr2 = ptr + 2

        # This will advance the pointer by 8 bytes (still pointer to long).
        self.assertEqual(ptr2.obj_offset, 8)
        self.assertEqual(ptr2, 0x33445566)

        # NOTE: We assume that long is 32 bits wide in both 64 bits and 32 bits
        # mode - which is the way windows does it. This is not the same as linux
        # which has long being the bit size in both cases.

        # Test the void pointer
        self.assertEqual(test.void, 8)

        # A Void object can not be compared to anything!
        self.assertNotEqual(test.void.dereference(), 0x33445566)

    def testArray(self):
        # Create an address space from a buffer for testing
        address_space = addrspace.BufferAddressSpace(
            data="abcdefghijklmnopqrstuvwxyz", session=self.session)

        profile = obj.Profile.classes['Profile32Bits'](session=self.session)
        test = profile.Object("Array", vm=address_space, offset=0,
                              target="int", count=0)

        self.assertEqual(test[0], 0x64636261)
        self.assertEqual(test[1], 0x68676665)

        # Can read past the end of the array but this returns all zeros.
        self.assertEqual(test[100], 0)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    unittest.main()



# Rekall Memory Forensics
# Copyright (C) 2011 Volatile Systems
# Copyright (C) 2011 Jamie Levy (Gleeda) <jamie.levy@gmail.com>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
@author:       Jamie Levy (Gleeda)
@license:      GNU General Public License 2.0 or later
@contact:      jamie.levy@gmail.com
@organization: Volatile Systems
"""

from rekall import utils
from rekall.plugins.windows.registry import registry

import hashlib
import struct


class GetServiceSids(registry.RegistryPlugin):
    """Get the names of services in the Registry and return Calculated SID"""

    __name = "getservicesids"

    def createservicesid(self, service_name):
        """Calculate the Service SID."""

        # We depend on service name to be a unicode string here.
        service_name = utils.SmartUnicode(service_name)

        sha = hashlib.sha1(service_name.encode("utf-16-le").upper()).digest()
        return 'S-1-5-80-' + '-'.join(
            [str(n) for n in struct.unpack("<IIIII", sha)])

    def get_service_sids(self):
        # Search for the current_control_set in all hives.
        for hive_offset in self.hive_offsets:
            reg = registry.RegistryHive(
                hive_offset=hive_offset, session=self.session)

            current_control_set = reg.CurrentControlSet()

            # There is no CurrentControlSet in this hive.
            if not current_control_set:
                continue

            # Enumerate the services.
            for subkey in current_control_set.open_subkey("services").subkeys():
                sid = self.createservicesid(subkey.Name)

                yield sid, subkey.Name

    def render(self, renderer):
        """output to Service SIDs as a dictionary for future use."""
        renderer.table_header([("SID", "sid", "<70"),
                               ("Service Name", "name", "")])

        for sid, service in self.get_service_sids():
            renderer.table_row(sid, service)



# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@google.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""This file implements an xls renderer based on the openpyxl project.

We produce xls (Excel spreadsheet files) with the output from Rekall plugins.
"""
import time
import openpyxl

from openpyxl import styles
from openpyxl.styles import colors
from openpyxl.styles import fills

from rekall import utils
from rekall.ui import renderer
from rekall.ui import text

# pylint: disable=unexpected-keyword-arg,no-value-for-parameter
# pylint: disable=redefined-outer-name

HEADER_STYLE = styles.Style(font=styles.Font(bold=True))
SECTION_STYLE = styles.Style(
    fill=styles.PatternFill(
        fill_type=fills.FILL_SOLID, start_color=styles.Color(colors.RED)))
FORMAT_STYLE = styles.Style(
    alignment=styles.Alignment(vertical="top", wrap_text=False))


class XLSObjectRenderer(renderer.ObjectRenderer):
    """By default the XLS renderer delegates to the text renderer."""
    renders_type = "object"
    renderers = ["XLSRenderer"]

    STYLE = None

    def _GetDelegateObjectRenderer(self, item):
        return self.ForTarget(item, "TextRenderer")(
            session=self.session, renderer=self.renderer.delegate_text_renderer)

    def RenderHeader(self, worksheet, column):
        cell = worksheet.cell(
            row=worksheet.current_row, column=worksheet.current_column)

        cell.value = column.name
        cell.style = HEADER_STYLE

        # Advance the pointer by 1 cell.
        worksheet.current_column += 1

    def RenderCell(self, value, worksheet, **options):
        # By default just render a single value into the current cell.
        cell = worksheet.cell(
            row=worksheet.current_row, column=worksheet.current_column)
        cell.value = self.GetData(value, **options)
        if self.STYLE:
            cell.style = self.STYLE

        # Advance the pointer by 1 cell.
        worksheet.current_column += 1

    def GetData(self, value, **options):
        if isinstance(value, (int, float, long)):
            return value

        return unicode(self._GetDelegateObjectRenderer(value).render_row(
            value, **options))


class XLSColumn(text.TextColumn):

    def __init__(self, type=None, table=None, renderer=None, session=None,
                 **options):
        super(XLSColumn, self).__init__(table=table, renderer=renderer,
                                        session=session, **options)

        if type:
            self.object_renderer = self.renderer.get_object_renderer(
                type=type, target_renderer="XLSRenderer", **options)


class XLSTable(text.TextTable):
    column_class = XLSColumn

    def render_header(self):
        current_ws = self.renderer.current_ws
        for column in self.columns:
            if column.object_renderer:
                object_renderer = column.object_renderer
            else:
                object_renderer = XLSObjectRenderer(
                    session=self.session, renderer=self.renderer)

            object_renderer.RenderHeader(self.renderer.current_ws, column)

        current_ws.current_row += 1
        current_ws.current_column = 1

    def render_row(self, row=None, highlight=None, **options):
        merged_opts = self.options.copy()
        merged_opts.update(options)

        # Get each column to write its own header.
        current_ws = self.renderer.current_ws
        for item in row:
            # Get the object renderer for the item.
            object_renderer = self.renderer.get_object_renderer(
                item, type=merged_opts.get("type"), **merged_opts)

            object_renderer.RenderCell(item, current_ws, **options)

        current_ws.current_row += 1
        current_ws.current_column = 1


class XLSRenderer(renderer.BaseRenderer):
    """A Renderer for xls files."""

    name = "xls"

    table_class = XLSTable
    tablesep = ""

    def __init__(self, output=None, **kwargs):
        super(XLSRenderer, self).__init__(**kwargs)

        # Make a single delegate text renderer for reuse. Most of the time we
        # will just replicate the output from the TextRenderer inside the
        # spreadsheet cell.
        self.delegate_text_renderer = text.TextRenderer(session=self.session)

        self.output = output or self.session.GetParameter("output")

        # If no output filename was give, just make a name based on the time
        # stamp.
        if self.output == None:
            self.output = "%s.xls" % time.ctime()

        try:
            self.wb = openpyxl.load_workbook(self.output)
            self.current_ws = self.wb.create_sheet()
        except IOError:
            self.wb = openpyxl.Workbook()
            self.current_ws = self.wb.active

    def start(self, plugin_name=None, kwargs=None):
        super(XLSRenderer, self).start(plugin_name=plugin_name, kwargs=kwargs)

        # Make a new worksheet for this run.
        if self.current_ws is None:
            self.current_ws = self.wb.create_sheet()

        ws = self.current_ws
        ws.title = plugin_name or ""
        ws.current_row = 1
        ws.current_column = 1

        return self

    def flush(self):
        super(XLSRenderer, self).flush()
        self.current_ws = None
        # Write the spreadsheet to a file.
        self.wb.save(self.output)

    def section(self, name=None, **_):
        ws = self.current_ws
        for i in range(10):
            cell = ws.cell(row=ws.current_row, column=i + 1)
            if i == 0:
                cell.value = name

            cell.style = SECTION_STYLE

        ws.current_row += 1
        ws.current_column = 1

    def format(self, formatstring, *data):
        worksheet = self.current_ws
        if "%" in formatstring:
            data = formatstring % data
        else:
            data = formatstring.format(*data)

        cell = worksheet.cell(
            row=worksheet.current_row, column=worksheet.current_column)
        cell.value = data
        cell.style = FORMAT_STYLE

        worksheet.current_column += 1
        if "\n" in data:
            worksheet.current_row += 1
            worksheet.current_column = 1

    def table_header(self, *args, **options):
        super(XLSRenderer, self).table_header(*args, **options)

        self.table.render_header()


# Following here are object specific renderers.


class XLSEProcessRenderer(XLSObjectRenderer):
    """Expands an EPROCESS into three columns (address, name and PID)."""
    renders_type = "_EPROCESS"

    def RenderHeader(self, worksheet, column):
        for heading in ["_EPROCESS", "Name", "PID"]:
            cell = worksheet.cell(
                row=worksheet.current_row, column=worksheet.current_column)
            cell.value = heading
            cell.style = HEADER_STYLE

            worksheet.current_column += 1

    def RenderCell(self, item, worksheet, **options):
        for value in ["%#x" % item.obj_offset, item.name, item.pid]:
            object_renderer = self.ForTarget(value, self.renderer)(
                session=self.session, renderer=self.renderer, **options)
            object_renderer.RenderCell(value, worksheet, **options)


class XLSStringRenderer(XLSObjectRenderer):
    renders_type = "String"

    def GetData(self, item, **_):
        return utils.SmartStr(item)


class XLSStructRenderer(XLSObjectRenderer):
    """Hex format struct's offsets."""
    renders_type = "Struct"

    def GetData(self, item, **_):
        return "%#x" % item.obj_offset


class XLSPointerRenderer(XLSObjectRenderer):
    """Renders the address of the pointer target as a hex string."""
    renders_type = "Pointer"

    def GetData(self, item, **_):
        result = item.v()
        if result == None:
            return "-"

        return "%#x" % result


class XLSNativeTypeRenderer(XLSObjectRenderer):
    """Renders native types as python objects."""
    renders_type = "NativeType"

    def GetData(self, item, **options):
        result = item.v()
        if result != None:
            return result


class XLS_UNICODE_STRING_Renderer(XLSNativeTypeRenderer):
    renders_type = "_UNICODE_STRING"


class XLSNoneObjectRenderer(XLSObjectRenderer):
    renders_type = "NoneObject"

    def GetData(self, item, **_):
        _ = item
        return "-"


class XLSDateTimeRenderer(XLSObjectRenderer):
    """Renders timestamps as python datetime objects."""
    renders_type = "UnixTimeStamp"
    STYLE = styles.Style(number_format='MM/DD/YYYY HH:MM:SS')

    def GetData(self, item, **options):
        if item.v() == 0:
            return None

        return item.as_datetime()



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
The Rekall Entity Layer.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

import re

from efilter import expression
from efilter import engine

from efilter.protocols import associative
from efilter.protocols import superposition


class ObjectMatcher(engine.VisitorEngine):
    """Given a query and bindings will evaluate the query."""

    # run() sets this to the sort value for the latest object matched.
    latest_sort_order = None

    def run(self, bindings, match_backtrace=False):
        self.match_backtrace = match_backtrace
        self.bindings = bindings

        # The match backtrace works by keeping a list of all the branches that
        # matched and then backtracking from the latest one to be evaluated
        # to the first parent that's a relation.
        if self.match_backtrace:
            self._matched_expressions = []

        self.latest_sort_order = []
        self.node = self.query.root
        self.result = self.visit(self.node)
        self.latest_sort_order = tuple(self.latest_sort_order)

        if self.match_backtrace:
            self.matched_expression = None
            for expr in self._matched_expressions:
                if isinstance(expr, expression.Relation):
                    self.matched_expression = expr
                    break

        if self.result:
            return self

        return False

    def visit(self, expr):
        result = super(ObjectMatcher, self).visit(expr)

        if self.match_backtrace and result:
            self._matched_expressions.append(expr)

        return result

    def visit_Literal(self, expr):
        return expr.value

    def visit_Binding(self, expr):
        return associative.select(self.bindings, expr.value)

    def visit_Let(self, expr):
        saved_bindings = self.bindings
        if isinstance(expr, expression.LetAny):
            union_semantics = True
        elif isinstance(expr, expression.LetEach):
            union_semantics = False
        else:
            union_semantics = None

        if not isinstance(expr.context, expression.Binding):
            raise ValueError(
                "Left operand of Let must be a Binding expression.")

        # Context to rebind to. This is the key that will be selected from
        # current bindings and become the new bindings for ever subexpression.
        context = expr.context.value

        try:
            rebind = associative.resolve(saved_bindings, context)

            if not rebind:  # No value from context.
                return None

            if union_semantics is None:
                # This is a simple let, which does not permit superposition
                # semantics.
                if superposition.insuperposition(rebind):
                    raise TypeError(
                        "A Let expression doesn't permit superposition "
                        "semantics. Use LetEach or LetAny instead.")

                self.bindings = rebind
                return self.visit(expr.expression)

            # If we're using union or intersection semantics, the type of
            # rebind MUST be a Superposition, even if it happens to have
            # only one state. If the below throws a type error then the
            # query is invalid and should fail here.
            result = False
            for state in superposition.getstates(rebind):
                self.bindings = state
                result = self.visit(expr.expression)
                if result and union_semantics:
                    return result

                if not result and not union_semantics:
                    return False

            return result
        finally:
            self.bindings = saved_bindings

    def visit_ComponentLiteral(self, expr):
        return getattr(self.bindings.components, expr.value)

    def visit_Complement(self, expr):
        return not self.visit(expr.value)

    def visit_Intersection(self, expr):
        for child in expr.children:
            if not self.visit(child):
                return False

        return True

    def visit_Union(self, expr):
        for child in expr.children:
            if self.visit(child):
                return True

        return False

    def visit_Sum(self, expr):
        return sum([self.visit(child) for child in expr.children])

    def visit_Difference(self, expr):
        difference = self.visit(expr.children[0])
        for child in expr.children[1:]:
            difference -= self.visit(child)

        return difference

    def visit_Product(self, expr):
        product = 1
        for child in expr.children:
            product *= self.visit(child)

        return product

    def visit_Quotient(self, expr):
        quotient = self.visit(expr.children[0])
        for child in expr.children[1:]:
            quotient /= self.visit(child)

        return quotient

    def visit_Equivalence(self, expr):
        first_val = self.visit(expr.children[0])
        for child in expr.children[1:]:
            if self.visit(child) != first_val:
                return False

        return True

    def visit_Membership(self, expr):
        return self.visit(expr.element) in set(self.visit(expr.set))

    def visit_RegexFilter(self, expr):
        string = self.visit(expr.string)
        pattern = self.visit(expr.regex)

        return re.compile(pattern).match(str(string))

    def visit_StrictOrderedSet(self, expr):
        iterator = iter(expr.children)
        min_ = self.visit(next(iterator))

        if min_ is None:
            return False

        for child in iterator:
            val = self.visit(child)

            if not min_ > val or val is None:
                return False

            min_ = val

        return True

    def visit_PartialOrderedSet(self, expr):
        iterator = iter(expr.children)
        min_ = self.visit(next(iterator))

        if min_ is None:
            return False

        for child in iterator:
            val = self.visit(child)
            if min_ < val or val is None:
                return False

            min_ = val

        return True

engine.Engine.register_engine(ObjectMatcher, "matcher")



# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""This address space overlays a pagefile into the physical address space.

This essentially implements the --pagefile parameter. Note that for images taken
with winpmem there is no need to specify the pagefile specifically since it is
already detected by the Elf64CoreDump class.
"""

__author__ = "Michael Cohen <scudette@gmail.com>"

from rekall import addrspace
from rekall import config
from rekall import session


config.DeclareOption(
    "--pagefile", type="ArrayStringParser", default=[],
    help="A pagefile to load into the image.")


class PagefilePhysicalAddressSpace(addrspace.MultiRunBasedAddressSpace):
    __image = True
    name = "pagefile"
    order = 200

    def __init__(self, **kwargs):
        super(PagefilePhysicalAddressSpace, self).__init__(**kwargs)
        pagefile_names = self.session.GetParameter("pagefile")

        self.as_assert(pagefile_names, "Pagefile not specified")
        self.as_assert(self.base.__class__ is not self.__class__)

        length = vaddr = 0
        for vaddr, _, length in self.base.get_available_addresses():
            self.add_run(vaddr, vaddr, length, self.base)

        vaddr += length + 0x10000

        # FIXME: Properly support multiple pagefiles.
        load_as = self.session.plugins.load_as(session=session.Session())
        for pagefile_name in pagefile_names:
            pagefile_as = load_as.GuessAddressSpace(filename=pagefile_name)

            if pagefile_as:
                self.pagefile_offset = vaddr
                vaddr += pagefile_as.end()
                self.pagefile_end = vaddr
                self.add_run(vaddr, 0, pagefile_as.end(), pagefile_as)



from rekall import testlib

from efilter import query as entity_query


class MatcherTest(testlib.RekallBaseUnitTestCase):
    def assertQueryMatches(self, query, bindings):
        m = entity_query.Query(query).run_engine("matcher", bindings=bindings)
        self.assertIsNotNone(m)

    def testBasic(self):
        query = "Process/pid is 1"
        bindings = {"Process/pid": 1}
        self.assertQueryMatches(query, bindings)

    def testRecursion(self):
        bindings = {"Process/parent": {"Process/pid": 1}}

        query = "Process/parent matches Process/pid is 1"
        self.assertQueryMatches(query, bindings)

        query = ("any Process/parent matches (Process/pid is 1 or "
                 "Process/command is 'foo')")
        self.assertQueryMatches(query, bindings)



# pylint: disable=unused-import

from rekall import obj_test
from rekall import addrspace_test

from rekall.plugins import tests

# Please do not move these to the rekall.entities namespace as that will be
# split off and no new dependencies on Rekall should be added.
from rekall.entities import entity_test
from rekall.entities import identity_test



# Rekall Memory Forensics
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
Darwin heap & VAD collectors.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

from rekall import obj

from rekall.entities import definitions

from rekall.plugins.collectors.darwin import common


class DarwinProcessVADs(common.DarwinEntityCollector):
    """Produces process virtual memory descriptors.

    Works by basically walking the vm_map_entry structs in each proc struct's
    task map, and making lots of inferences based on the values in that and
    related structures.

    Most of the actual derivation logic is in the overlay.
    """

    outputs = ["MemoryDescriptor", "AddressSpace", "Struct/type=vm_map_entry"]

    collect_args = dict(procs="has component Process")
    enforce_hint = True

    PROTECTION_FLAGS = {
        "VM_PROT_READ": "read",
        "VM_PROT_WRITE": "write",
        "VM_PROT_EXECUTE": "execute"}

    def collect(self, hint, procs):
        # We're going to have a much better time if we can ignore processes
        # that aren't relevant.
        hint_filter = hint.run_engine("hinter",
                                      selector="MemoryDescriptor/process")

        for proc in procs:
            if (hint_filter and not
                    hint_filter.run_engine("matcher", bindings=proc)):
                continue

            vm = proc["Struct/base"].get_process_address_space()
            arch = "x86_32" if vm.end() <= 0x1000000000 else "x86_64"

            vm_identity, vm_entity = self.prebuild(
                components=[definitions.AddressSpace(dtb=vm.dtb,
                                                     type="virtual",
                                                     architecture=arch,
                                                     owner=proc.identity),
                            definitions.Named(kind="Virtual Address Space",
                                              name="DTB @%#016x" % vm.dtb)],
                keys=("AddressSpace/dtb",))

            yield vm_entity

            for address in proc["Struct/base"].task.map.hdr.walk_list(
                    "links.next", include_current=False):

                # Is this range backed by a vnode? If so yield it, and save the
                # identity. This should take care of __TEXT, __LINKEDIT and
                # memory mapped sections.
                vnode = address.find_vnode_object()
                file_identity = None
                if vnode:
                    file_identity, file = self.prebuild(
                        components=[definitions.Struct(base=vnode,
                                                       type="vnode")],
                        keys=("Struct/base",))
                    yield file

                # Sort out permission bits.
                perms = []
                max_perms = []
                for flag in address.protection:
                    perm = self.PROTECTION_FLAGS.get(flag)
                    if not perm:
                        raise ValueError("Unknown protection flag %r." % flag)
                    perms.append(perm)

                for flag in address.max_protection:
                    perm = self.PROTECTION_FLAGS.get(flag)
                    if not perm:
                        raise ValueError("Unknown protection flag %r." % flag)
                    max_perms.append(perm)

                # Sort out shared memory flags.
                sharing_mode = address.sharing_mode
                sharing_enum = None
                if sharing_mode in ("SM_SHARED", "SM_LARGE_PAGE"):
                    sharing_enum = "shared"
                elif sharing_mode == "SM_COW":
                    sharing_enum = "copy-on-write"
                elif sharing_mode == "SM_PRIVATE":
                    sharing_enum = "private"
                elif sharing_mode == "SM_EMPTY":
                    sharing_enum = None
                else:
                    raise ValueError("Unknown sharing mode %r." % sharing_mode)

                _, result = self.prebuild(
                    components=[
                        definitions.MemoryDescriptor(
                            start=obj.Void(vm=vm,
                                           profile=self.session.profile,
                                           offset=address.links.start.v()),
                            end=obj.Void(vm=vm,
                                         profile=self.session.profile,
                                         offset=address.links.end.v()),
                            address_space=vm_identity,
                            permissions=perms,
                            max_permissions=max_perms,
                            file=file_identity,
                            shared=sharing_enum,
                            code_signed=address.code_signed),
                        definitions.Struct(
                            base=address,
                            type="vm_map_entry")],
                    keys=("MemoryDescriptor/address_space",
                          "MemoryDescriptor/start",
                          "MemoryDescriptor/end"))

                yield result



#!/usr/bin/env python

# Rekall
# Copyright (C) 2012 Michael Cohen <scudette@gmail.com>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

__author__ = "Michael Cohen <scudette@gmail.com>"

# pylint: disable=protected-access

import logging
import pdb
import sys


from rekall import args
from rekall import config
from rekall import plugin
from rekall import session
from rekall.ui import text

from pkg_resources import iter_entry_points
for entry_point in iter_entry_points(group='rekall.plugins', name=None):
    entry_point.load()

from rekall import plugins


class Run(plugin.Command):
    """A plugin which runs its argument (using eval).

    Note: This plugin is only defined and available when using the main entry
    point. It is not available when Rekall is used as a library since it allows
    arbitrary code execution.
    """

    name = "run"

    @classmethod
    def args(cls, parser):
        super(Run, cls).args(parser)
        parser.add_argument("script", default="print 'hello!'",
                            help="The script to evaluate")

        parser.add_argument("--run", default=None,
                            help="A file name to run.")

    def __init__(self, script, run=None, **kwargs):
        super(Run, self).__init__(**kwargs)
        if run is not None:
            script = open(run).read()

        exec script in self.session.locals


def main(argv=None):
    # New user interactive session (with extra bells and whistles).
    user_session = session.InteractiveSession()
    user_session.session_list.append(user_session)
    text_renderer = text.TextRenderer(session=user_session)

    with text_renderer.start():
        plugin_cls, flags = args.parse_args(argv=argv,
                                            user_session=user_session)

    try:
        # Run the plugin with plugin specific args.
        user_session.RunPlugin(plugin_cls, **config.RemoveGlobalOptions(flags))
    except Exception as e:
        logging.fatal("%s. Try --debug for more information." % e)
        if getattr(flags, "debug", None):
            pdb.post_mortem(sys.exc_info()[2])
        raise
    finally:
        user_session.Flush()


if __name__ == '__main__':
    main()



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
The Rekall Entity Layer.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter.protocols import associative
from efilter.protocols import indexable
from efilter.protocols import hashable
from efilter.protocols import name_delegate

from rekall import obj
from rekall import plugin as rekall_plugin

from rekall.entities import entity
from rekall.entities import identity
from rekall.entities import component as entity_component


class RekallDelegate(object):
    """Exposes the global Rekall namespace and types to EFILTER.

    This is a work in progress.
    """

    def __init__(self, session, profile):
        self.session = session
        self.profile = profile

    def _reflect_global(self, name):
        if name in entity_component.Component.classes.keys():
            # Is this a valid component? If so, just tell EFILTER it's an
            # Entity.
            return entity_component.Component.classes.get(name)
        elif name.startswith("_"):
            # Could be a global.
            value = self.profile.get_constant(name)
            if value:
                return type(value)

            return None
        else:
            # Try a plugin name.
            # If name is in session plugins, return the plugin class.
            return None

    def _reflect_scoped(self, name, scope):
        if issubclass(scope, entity_component.Component):
            return self._reflect_component(name, scope)

        if issubclass(scope, obj.BaseObject):
            return self._reflect_vtype(name, scope)

        if issubclass(scope, rekall_plugin.Plugin):
            return self._reflect_plugin(name, scope)

        return None

    def _reflect_component(self, name, component):
        field = component.reflect_attribute(name)
        if not field:
            return None

        return getattr(field.typedesc, "type_cls", None)

    def _reflect_vtype(self, name, vtype):
        pass

    def _reflect_plugin(self, name, plugin):
        pass

    def reflect(self, name, scope=None):
        if scope is None:
            return self._reflect_global(name)

        return self._reflect_scoped(name, scope)

    def provide(self, name):
        return None

    def getnames(self):
        return ()


name_delegate.INameDelegate.implement(
    for_type=RekallDelegate,
    implementations={
        name_delegate.reflect: RekallDelegate.reflect,
        name_delegate.provide: RekallDelegate.provide,
        name_delegate.getnames: RekallDelegate.getnames})


def _getkeys_Entity(e):
    for component_name in entity_component.Component.classes.keys():
        component = getattr(e.components, component_name)
        if component is None:
            continue

        for idx, field in enumerate(component.component_fields):
            if component[idx]:
                yield "%s/%s" % (component_name, field.name)


### Entity-related types: ###

associative.IAssociative.implement(
    for_type=entity.Entity,
    implementations={
        associative.select: lambda e, key: e.get_raw(key),
        associative.resolve: lambda e, key: e.get(key),
        associative.getkeys: _getkeys_Entity})


associative.IAssociative.implement(
    for_type=entity_component.Component,
    implementations={
        associative.select: lambda c, key: c[key],
        associative.resolve: lambda c, key: c[key],
        associative.getkeys: lambda c: (f.name for f in c.component_fields)})


associative.IAssociative.implement(
    for_type=entity.CurriedComponent,
    implementations={
        associative.select: entity.CurriedComponent.get,
        associative.resolve: entity.CurriedComponent.get_raw,
        associative.getkeys:
            lambda c: (f.name for f in c.component.component_fields)})


indexable.IIndexable.implement(
    for_types=(identity.Identity, entity.Entity),
    implementations={
        indexable.indices: lambda x: x.indices})

### Structs/vtypes: ###

associative.IAssociative.implement(
    for_type=obj.Struct,
    implementations={
        associative.select: lambda o, key: o.m(key),
        associative.resolve: lambda o, key: getattr(o, key, None),
        associative.getkeys: lambda o: o.members.iterkeys()})


indexable.IIndexable.implement(
    for_type=obj.Struct,
    implementations={
        indexable.indices: lambda o: o.indices})


hashable.IHashable.implement(
    for_type=obj.BaseObject,
    implementations={
        hashable.hashed: hash})



# Rekall Memory Forensics
# Copyright (C) 2007-2013 Volatility Foundation
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This file is part of Rekall Memory Forensics.
#
# Rekall Memory Forensics is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General Public
# License.
#
# Rekall Memory Forensics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Rekall Memory Forensics.  If not, see <http://www.gnu.org/licenses/>.
#

"""
@author:       Andrew Case
@license:      GNU General Public License 2.0
@contact:      atcuno@gmail.com
@organization:
"""
from rekall.plugins.linux import common

class PSAux(common.LinProcessFilter):
    """Gathers processes along with full command line and start time."""

    __name = "psaux"

    def render(self, renderer):
        renderer.table_header([
                ("PID", "pid", "5"),
                ("UID", "uid", "5"),
                ("GID", "gid", "5"),
                ("Command", "command", "50")])

        for task in self.filter_processes():
            renderer.table_row(task.pid, task.uid, task.gid, task.commandline)
